# SPDX-License-Identifier: MIT

from datetime import datetime
import sqlite3
import os

from amd_debug.common import read_file

SCHEMA_VERSION = 1


def migrate(cur, user_version) -> None:
    """Migrate sqlite database schema"""
    cur.execute("PRAGMA user_version")
    val = cur.fetchone()[0]
    # Schema 1
    # - add priority column
    if val == 0:
        cur.execute("ALTER TABLE debug ADD COLUMN priority INTEGER")
    # Update schema if necessary
    if val != user_version:
        cur.execute(f"PRAGMA user_version = {user_version}")


class SleepDatabase:
    """Database class to store sleep cycle data"""

    def __init__(self, dbf=None) -> None:
        self.db = None
        self.last_suspend = None
        self.cycle_data_cnt = 0
        self.debug_cnt = 0

        if not dbf:
            # if we were packaged we would have a directory in /var/lib
            path = os.path.join("/", "var", "lib", "amd-s2idle")
            if not os.path.exists(path):
                path = os.path.join("/", "var", "local", "lib", "amd-s2idle")
            os.makedirs(path, exist_ok=True)

            dbf = os.path.join(path, "data.db")
        new = not os.path.exists(dbf)
        self.db = sqlite3.connect(dbf)
        cur = self.db.cursor()
        cur.execute(
            "CREATE TABLE IF NOT EXISTS prereq_data ("
            "t0 INTEGER,"
            "id INTEGER,"
            "message TEXT,"
            "symbol TEXT,"
            "PRIMARY KEY(t0, id))"
        )
        cur.execute(
            "CREATE TABLE IF NOT EXISTS debug ("
            "t0 INTEGER,"
            "id INTEGER,"
            "message TEXT,"
            "priority INTEGER,"
            "PRIMARY KEY(t0, id))"
        )
        cur.execute(
            "CREATE TABLE IF NOT EXISTS cycle ("
            "t0 INTEGER PRIMARY KEY,"
            "t1 INTEGER,"
            "requested INTEGER,"
            "gpio TEXT,"
            "wake_irq TEXT,"
            "kernel REAL,"
            "hw REAL)"
        )
        cur.execute(
            "CREATE TABLE IF NOT EXISTS cycle_data ("
            "t0 INTEGER,"
            "id INTEGER,"
            "message TEXT,"
            "symbol TEXT,"
            "PRIMARY KEY(t0, id))"
        )
        cur.execute(
            "CREATE TABLE IF NOT EXISTS battery ("
            "t0 INTEGER PRIMARY KEY,"
            "name TEXT,"
            "b0 INTEGER,"
            "b1 INTEGER,"
            "full INTEGER,"
            "unit TEXT)"
        )
        self.prereq_data_cnt = 0

        if new:
            cur.execute(f"PRAGMA user_version = {SCHEMA_VERSION}")
        else:
            migrate(cur, SCHEMA_VERSION)

    def __del__(self) -> None:
        if self.db:
            self.db.close()

    def start_cycle(self, timestamp):
        """Start a new sleep cycle"""
        assert self.db
        self.last_suspend = timestamp
        # increment the counters so that systemd hooks work
        cur = self.db.cursor()
        cur.execute(
            "SELECT MAX(id) FROM cycle_data WHERE t0=?",
            (int(self.last_suspend.strftime("%Y%m%d%H%M%S")),),
        )
        val = cur.fetchone()[0]
        if val is not None:
            self.cycle_data_cnt = val + 1
        else:
            self.cycle_data_cnt = 0
        cur.execute(
            "SELECT MAX(id) FROM debug WHERE t0=?",
            (int(self.last_suspend.strftime("%Y%m%d%H%M%S")),),
        )
        val = cur.fetchone()[0]
        if val is not None:
            self.debug_cnt = val + 1
        else:
            self.debug_cnt = 0

    def sync(self) -> None:
        """Sync the database to disk"""
        assert self.db
        self.db.commit()

    def record_debug(self, message, level=6) -> None:
        """Helper function to record a message to debug database"""
        assert self.last_suspend
        assert self.db
        cur = self.db.cursor()
        cur.execute(
            "INSERT into debug (t0, id, message, priority) VALUES (?, ?, ?, ?)",
            (
                int(self.last_suspend.strftime("%Y%m%d%H%M%S")),
                self.debug_cnt,
                message,
                level,
            ),
        )
        self.debug_cnt += 1

    def record_debug_file(self, fn):
        """Helper function to record the entire contents of a file to debug database"""
        try:
            contents = read_file(fn)
            self.record_debug(contents)
        except PermissionError:
            self.record_debug(f"Unable to capture {fn}")

    def record_battery_energy(self, name, energy, full, unit):
        """Helper function to record battery energy"""
        assert self.db
        assert self.last_suspend
        cur = self.db.cursor()
        cur.execute(
            "SELECT * FROM battery WHERE t0=?",
            (int(self.last_suspend.strftime("%Y%m%d%H%M%S")),),
        )
        if cur.fetchone():
            cur.execute(
                "UPDATE battery SET b1=? WHERE t0=?",
                (energy, int(self.last_suspend.strftime("%Y%m%d%H%M%S"))),
            )
        else:
            cur.execute(
                """
                INSERT into battery (t0, name, b0, b1, full, unit) 
                VALUES (?, ?, ?, ?, ?, ?)
                """,
                (
                    int(self.last_suspend.strftime("%Y%m%d%H%M%S")),
                    name,
                    energy,
                    None,
                    full,
                    unit,
                ),
            )

    def record_cycle_data(self, message, symbol) -> None:
        """Helper function to record a message to cycle_data database"""
        assert self.last_suspend
        assert self.db
        cur = self.db.cursor()
        cur.execute(
            """
            INSERT into cycle_data (t0, id, message, symbol)
            VALUES (?, ?, ?, ?)
            """,
            (
                (
                    int(self.last_suspend.strftime("%Y%m%d%H%M%S")),
                    self.cycle_data_cnt,
                    message,
                    symbol,
                )
            ),
        )
        self.cycle_data_cnt += 1

    def record_cycle(
        self,
        requested_duration=0,
        active_gpios="",
        wakeup_irqs="",
        kernel_duration=0,
        hw_sleep_duration=0,
    ) -> None:
        """Helper function to record a sleep cycle into the cycle database"""
        assert self.last_suspend
        assert self.db
        cur = self.db.cursor()
        cur.execute(
            """
            REPLACE INTO cycle (t0, t1, requested, gpio, wake_irq, kernel, hw)
            VALUES (?, ?, ?, ?, ?, ?, ?)
            """,
            (
                int(self.last_suspend.strftime("%Y%m%d%H%M%S")),
                int(datetime.now().strftime("%Y%m%d%H%M%S")),
                requested_duration,
                str(active_gpios) if active_gpios else "",
                str(wakeup_irqs),
                kernel_duration,
                hw_sleep_duration,
            ),
        )

    def record_prereq(self, message, symbol) -> None:
        """Helper function to record a message to prereq_data database"""
        assert self.last_suspend
        assert self.db
        cur = self.db.cursor()
        cur.execute(
            """
            INSERT into prereq_data (t0, id, message, symbol)
            VALUES (?, ?, ?, ?)
            """,
            (
                (
                    int(self.last_suspend.strftime("%Y%m%d%H%M%S")),
                    self.prereq_data_cnt,
                    message,
                    symbol,
                )
            ),
        )
        self.prereq_data_cnt += 1

    def report_prereq(self, t0) -> list:
        """Helper function to report the prereq_data database"""
        assert self.db
        if t0 is None:
            return []
        cur = self.db.cursor()
        cur.execute(
            "SELECT * FROM prereq_data WHERE t0=?",
            (int(t0.strftime("%Y%m%d%H%M%S")),),
        )
        return cur.fetchall()

    def report_debug(self, t0) -> list:
        """Helper function to report the debug database"""
        assert self.db
        if t0 is None:
            return []
        cur = self.db.cursor()
        cur.execute(
            "SELECT message, priority FROM debug WHERE t0=?",
            (int(t0.strftime("%Y%m%d%H%M%S")),),
        )
        return cur.fetchall()

    def report_cycle(self, t0=None) -> list:
        """Helper function to report a cycle from database"""
        assert self.db
        if t0 is None:
            assert self.last_suspend
            t0 = self.last_suspend
        cur = self.db.cursor()
        cur.execute(
            "SELECT * FROM cycle WHERE t0=?",
            (int(t0.strftime("%Y%m%d%H%M%S")),),
        )
        return cur.fetchall()

    def report_cycle_data(self, t0=None) -> str:
        """Helper function to report a table matching a timestamp from cycle_data database"""
        assert self.db
        if t0 is None:
            assert self.last_suspend
            t0 = self.last_suspend
        cur = self.db.cursor()
        cur.execute(
            "SELECT message, symbol FROM cycle_data WHERE t0=? ORDER BY symbol",
            (int(t0.strftime("%Y%m%d%H%M%S")),),
        )
        data = ""
        for row in cur.fetchall():
            data += f"{row[1]} {row[0]}\n"
        return data

    def report_battery(self, t0=None) -> list:
        """Helper function to report a line from battery database"""
        assert self.db
        if t0 is None:
            assert self.last_suspend
            t0 = self.last_suspend
        cur = self.db.cursor()
        cur.execute(
            "SELECT * FROM battery WHERE t0=?",
            (int(t0.strftime("%Y%m%d%H%M%S")),),
        )
        return cur.fetchall()

    def get_last_prereq_ts(self) -> int:
        """Helper function to report the last line from prereq database"""
        assert self.db
        cur = self.db.cursor()
        cur.execute("SELECT * FROM prereq_data ORDER BY t0 DESC LIMIT 1")
        result = cur.fetchone()
        return result[0] if result else 0

    def get_last_cycle(self) -> list:
        """Helper function to report the last line from battery database"""
        assert self.db
        cur = self.db.cursor()
        cur.execute("SELECT t0 FROM cycle ORDER BY t0 DESC LIMIT 1")
        return cur.fetchone()

    def report_summary_dataframe(self, since, until) -> object:
        """Helper function to report a dataframe from the database"""
        assert self.db

        import pandas as pd  # pylint: disable=import-outside-toplevel

        pd.set_option("display.precision", 2)
        return pd.read_sql_query(
            sql="SELECT cycle.t0, cycle.t1, hw, requested, gpio, wake_irq, b0, b1, full FROM cycle LEFT JOIN battery ON cycle.t0 = battery.t0 WHERE cycle.t0 >= ? and cycle.t0 <= ?",
            con=self.db,
            params=(
                int(since.strftime("%Y%m%d%H%M%S")),
                int(until.strftime("%Y%m%d%H%M%S")),
            ),
        )



# SPDX-License-Identifier: MIT

import os
import logging
from amd_debug.common import BIT, read_file

ACPI_METHOD = "M460"


def search_acpi_tables(pattern):
    """Search for a pattern in ACPI tables"""
    p = os.path.join("/", "sys", "firmware", "acpi", "tables")

    for fn in os.listdir(p):
        if not fn.startswith("SSDT") and not fn.startswith("DSDT"):
            continue
        fp = os.path.join(p, fn)
        with open(fp, "rb") as file:
            content = file.read()
            if pattern.encode() in content:
                return True
    return False


class AcpicaTracer:
    """Class for ACPI tracing"""

    def __init__(self):
        self.acpi_base = os.path.join("/", "sys", "module", "acpi", "parameters")
        keys = [
            "trace_debug_layer",
            "trace_debug_level",
            "trace_method_name",
            "trace_state",
        ]
        self.original = {}
        self.supported = False
        for key in keys:
            fname = os.path.join(self.acpi_base, key)
            if not os.path.exists(fname):
                logging.debug("ACPI Notify() debugging not available")
                return
            v = read_file(fname)
            if v and v != "(null)":
                self.original[key] = v
        self.supported = True

    def _write_expected(self, expected):
        for key, value in expected.items():
            p = os.path.join(self.acpi_base, key)
            if isinstance(value, int):
                t = str(int(value))
            else:
                t = value
            with open(p, "w", encoding="utf-8") as w:
                w.write(t)

    def trace_notify(self):
        """Trace notify events"""
        if not self.supported:
            return False
        expected = {
            "trace_debug_layer": BIT(2),
            "trace_debug_level": BIT(2),
            "trace_state": "enable",
        }
        self._write_expected(expected)
        logging.debug("Enabled ACPI debugging for ACPI_LV_INFO/ACPI_EVENTS")
        return True

    def trace_bios(self):
        """Trace BIOS events"""
        if not self.supported:
            return False
        if not search_acpi_tables(ACPI_METHOD):
            logging.debug(
                "will not work on this system: ACPI tables do not contain %s",
                ACPI_METHOD,
            )
            return False
        expected = {
            "trace_debug_layer": BIT(7),
            "trace_debug_level": BIT(4),
            "trace_method_name": f"\\{ACPI_METHOD}",
            "trace_state": "method",
        }
        self._write_expected(expected)
        logging.debug("Enabled ACPI debugging for BIOS")
        return True

    def disable(self):
        """Disable ACPI tracing"""
        if not self.supported:
            return False
        expected = {
            "trace_state": "disable",
        }
        self._write_expected(expected)
        return True

    def restore(self):
        """Restore original ACPI tracing settings"""
        if not self.supported:
            return False
        self._write_expected(self.original)
        return True



#!/usr/bin/python3
# SPDX-License-Identifier: MIT

"""
This module contains unit tests for the kernel log functions in the amd-debug-tools package.
"""
from unittest.mock import patch, mock_open

import unittest
import logging

from amd_debug.kernel import sscanf_bios_args, get_kernel_command_line, DmesgLogger


class TestKernelLog(unittest.TestCase):
    """Test Common kernel scan functions"""

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(filename="/dev/null", level=logging.DEBUG)

    def test_get_kernel_command_line(self):
        """Test get_kernel_command_line function"""

        # Test case with a valid kernel command line that is fully filtered
        kernel_cmdline = "quiet splash"
        expected_output = ""
        with patch(
            "amd_debug.common.open", new_callable=mock_open, read_data=kernel_cmdline
        ) as _mock_file:
            result = get_kernel_command_line()
            self.assertEqual(result, expected_output)

        # Test case with an empty kernel command line
        kernel_cmdline = ""
        expected_output = ""
        with patch(
            "amd_debug.common.open", new_callable=mock_open, read_data=kernel_cmdline
        ) as _mock_file:
            result = get_kernel_command_line()
            self.assertEqual(result, expected_output)

        # Test case with a kernel command line containing special characters
        kernel_cmdline = "quiet splash --debug=1"
        expected_output = "--debug=1"
        with patch(
            "amd_debug.common.open", new_callable=mock_open, read_data=kernel_cmdline
        ) as _mock_file:
            result = get_kernel_command_line()
            self.assertEqual(result, expected_output)

        # Test case with a kernel command line containing special characters
        kernel_cmdline = "quiet splash initrd=foo modprobe.blacklist=foo"
        expected_output = "modprobe.blacklist=foo"
        with patch(
            "amd_debug.common.open", new_callable=mock_open, read_data=kernel_cmdline
        ) as _mock_file:
            result = get_kernel_command_line()
            self.assertEqual(result, expected_output)

    def test_sscanf_bios_args(self):
        """Test sscanf_bios_args function"""

        # Test case with a valid line
        line = 'ex_trace_args: "format_string", 0x1234, 0x5678'
        expected_output = "format_string"
        result = sscanf_bios_args(line)
        self.assertEqual(result, expected_output)

        # Test case with an invalid line
        line = 'invalid_line: "format_string", 0x1234, 0x5678'
        result = sscanf_bios_args(line)
        self.assertIsNone(result)

        # Test case with a line containing "Unknown"
        line = 'ex_trace_args: "format_string", Unknown, 0x5678'
        expected_output = "format_string"
        result = sscanf_bios_args(line)
        self.assertEqual(result, expected_output)

        # make sure that lines with ex_trace_point are not parsed
        line = 'ex_trace_point: "format_string", 0x1234, 0x5678'
        result = sscanf_bios_args(line)
        self.assertTrue(result)

        # test a real post code line
        line = 'ex_trace_args: "  POST CODE: %X  ACPI TIMER: %X  TIME: %d.%d ms\\n", b0003f33, 83528798, 0, 77, 0, 0'
        expected_output = "POST CODE: B0003F33  ACPI TIMER: 83528798  TIME: 0.119 ms"
        result = sscanf_bios_args(line)
        self.assertEqual(result, expected_output)

        # test a real _REG print
        line = 'ex_trace_args:  "  OEM-ASL-PCIe Address (0x%X)._REG (%d %d)  PCSA = %d\\n", ec303000, 2, 0, 0, 0, 0'
        expected_output = "OEM-ASL-PCIe Address (0xEC303000)._REG (2 0)  PCSA = 0"
        result = sscanf_bios_args(line)
        self.assertEqual(result, expected_output)

        # test case of too may arguments
        line = 'ex_trace_args         :  "  APGE                  = %d\\n", 1, 0, 0, 0, 0, 0'
        expected_output = "APGE                  = 1"
        result = sscanf_bios_args(line)
        self.assertEqual(result, expected_output)

        # test case for Dispatch notify
        line = "evmisc-0132 ev_queue_notify_reques: Dispatching Notify on [UBTC] (Device) Value 0x80 (Status Change) Node 00000000851b15c1"
        expected_output = (
            "Dispatching Notify on [UBTC] (Device) Value 0x80 (Status Change)"
        )
        result = sscanf_bios_args(line)
        self.assertEqual(result, expected_output)


class TestDmesgLogger(unittest.TestCase):
    """Test Dmesg logger functions"""

    @classmethod
    @patch("subprocess.run")
    def setUpClass(cls, _mock_run=None):
        logging.basicConfig(filename="/dev/null", level=logging.DEBUG)

    def test_dmesg_logger_initialization(self):
        """Test initialization of DmesgLogger"""

        with patch("subprocess.run") as mock_run:
            # Mock the subprocess output for dmesg -h
            mock_run.return_value.stdout = b"--since supported\n"
            mock_run.return_value.returncode = 0

            logger = DmesgLogger()
            self.assertTrue(logger.since_support)
            self.assertEqual(logger.command, ["dmesg", "-t", "-k"])

    def test_dmesg_logger_refresh_head(self):
        """Test _refresh_head method of DmesgLogger"""

        with patch("subprocess.run") as mock_run:
            # Mock the subprocess output for dmesg
            mock_run.return_value.stdout = b"line1\nline2\n"
            mock_run.return_value.returncode = 0

            logger = DmesgLogger()
            logger._refresh_head()  # pylint: disable=protected-access
            self.assertEqual(logger.buffer, "line1\nline2\n")

    def test_dmesg_logger_seek_tail(self):
        """Test seek_tail method of DmesgLogger"""

        with patch("subprocess.run") as mock_run:
            # Mock the subprocess output for dmesg
            mock_run.return_value.stdout = b"line1\nline2\n"
            mock_run.return_value.returncode = 0

            logger = DmesgLogger()
            logger.seek_tail()
            self.assertEqual(logger.buffer, "line1\nline2\n")

    def test_dmesg_logger_process_callback(self):
        """Test process_callback method of DmesgLogger"""

        with patch("subprocess.run") as mock_run:
            # Mock the subprocess output for dmesg
            mock_run.return_value.stdout = b"line1\nline2\n"
            mock_run.return_value.returncode = 0

            logger = DmesgLogger()
            logger._refresh_head()  # pylint: disable=protected-access

            mock_callback = unittest.mock.Mock()
            logger.process_callback(mock_callback)

            mock_callback.assert_any_call("line1", None)
            mock_callback.assert_any_call("line2", None)

    def test_dmesg_logger_match_line(self):
        """Test match_line method of DmesgLogger"""

        with patch("subprocess.run") as mock_run:
            # Mock the subprocess output for dmesg
            mock_run.return_value.stdout = b"line1\nline2\n"
            mock_run.return_value.returncode = 0

            logger = DmesgLogger()
            logger._refresh_head()  # pylint: disable=protected-access

            result = logger.match_line(["line1"])
            self.assertEqual(result, "line1")

            result = logger.match_line(["nonexistent"])
            self.assertEqual(result, "")

    def test_dmesg_logger_match_pattern(self):
        """Test match_pattern method of DmesgLogger"""

        with patch("subprocess.run") as mock_run:
            # Mock the subprocess output for dmesg
            mock_run.return_value.stdout = b"line1\nline2\n"
            mock_run.return_value.returncode = 0

            logger = DmesgLogger()
            logger._refresh_head()  # pylint: disable=protected-access

            result = logger.match_pattern(r"line\d")
            self.assertEqual(result, "line1")

            result = logger.match_pattern(r"nonexistent")
            self.assertEqual(result, "")



#!/usr/bin/python3
# SPDX-License-Identifier: MIT
"""
This module is a launcher for the AMD Debug Tools package. It is meant for
launching various tools within the package without installation.
"""

import sys
import os

URL = "git://git.kernel.org/pub/scm/linux/kernel/git/superm1/amd-debug-tools.git"
try:
    import amd_debug
    from amd_debug.common import fatal_error
except ModuleNotFoundError:
    sys.exit(
        f"\033[91m{sys.argv[0]} can not be run standalone.\n"
        f"\033[0m\033[94mCheck out the full branch from {URL}\033[0m"
    )


def main():
    """Main function to launch the appropriate tool based on the script name."""
    try:
        return amd_debug.launch_tool(os.path.basename(sys.argv[0]))
    except ModuleNotFoundError as e:
        fatal_error(
            f"Missing dependency: {e}\n"
            f"Run ./install_deps.py to install dependencies."
        )


if __name__ == "__main__":
    sys.exit(main())



# SPDX-License-Identifier: MIT
"""Display analysis"""
import os
from pyudev import Context

from amd_debug.common import read_file


class Display:
    """Display analysis"""

    def __init__(self):
        self.pyudev = Context()
        self.edid = []

        for dev in self.pyudev.list_devices(subsystem="drm"):
            if not "card" in dev.device_path:
                continue
            p = os.path.join(dev.sys_path, "status")
            if not os.path.exists(p):
                continue
            f = read_file(p)
            if f != "connected":
                continue
            p = os.path.join(dev.sys_path, "enabled")
            f = read_file(p)
            if f != "enabled":
                continue
            self.edid.append(os.path.join(dev.sys_path, "edid"))

    def get_edid(self) -> list:
        """Get the path for EDID data for all connected displays"""
        return self.edid



#!/usr/bin/python3
# SPDX-License-Identifier: MIT

"""
This module contains unit tests for the s2idle tool in the amd-debug-tools package.
"""

import math
import unittest
from datetime import datetime
from unittest.mock import patch
import pandas as pd

from amd_debug.sleep_report import (
    remove_duplicates,
    format_gpio_as_str,
    format_irq_as_str,
    format_as_human,
    format_as_seconds,
    format_watts,
    format_percent,
    format_timedelta,
    parse_hw_sleep,
    SleepReport,
)

from amd_debug.wake import WakeGPIO, WakeIRQ


class TestSleepReportUtils(unittest.TestCase):
    """Unit tests for the sleep report utilities."""

    def test_remove_duplicates(self):
        """Test the remove_duplicates function."""
        self.assertEqual(remove_duplicates("1, 2, 2, 3"), [1, 2, 3])
        self.assertEqual(remove_duplicates("4 4 5 6"), [4, 5, 6])
        self.assertEqual(remove_duplicates(""), [])

    def test_format_gpio_as_str(self):
        """Test the format_gpio_as_str function."""
        self.assertEqual(format_gpio_as_str("1, 2, 2, 3"), "1, 2, 3")
        self.assertEqual(format_gpio_as_str("4 4 5 6"), "4, 5, 6")
        self.assertEqual(format_gpio_as_str(""), "")

    @patch("amd_debug.wake.read_file")
    @patch("os.path.exists")
    @patch("os.listdir")
    @patch("os.walk")
    def test_format_irq_as_str(
        self, _mock_os_walk, _mock_os_listdir, mock_os_path_exists, mock_read_file
    ):
        """Test the format_irq_as_str function."""
        mock_read_file.side_effect = lambda path: {
            "/sys/kernel/irq/20/chip_name": "",
            "/sys/kernel/irq/20/actions": "",
            "/sys/kernel/irq/20/wakeup": "disabled",
        }.get(path, "")

        # Mocking os.path.exists
        mock_os_path_exists.return_value = False

        self.assertEqual(format_irq_as_str("20"), "Disabled interrupt")
        self.assertEqual(format_irq_as_str(""), "")

    def test_format_as_human(self):
        """Test the format_as_human function."""
        self.assertEqual(
            format_as_human("20231010123045"),
            datetime(2023, 10, 10, 12, 30, 45),
        )
        with self.assertRaises(ValueError):
            format_as_human("invalid_date")

    def test_format_as_seconds(self):
        """Test the format_as_seconds function."""
        self.assertEqual(
            format_as_seconds("20231010123045"),
            datetime(2023, 10, 10, 12, 30, 45).timestamp(),
        )
        with self.assertRaises(ValueError):
            format_as_seconds("invalid_date")

    def test_format_watts(self):
        """Test the format_watts function."""
        self.assertEqual(format_watts(12.3456), "12.35W")
        self.assertEqual(format_watts(0), "0.00W")

    def test_format_percent(self):
        """Test the format_percent function."""
        self.assertEqual(format_percent(12.3456), "12.35%")
        self.assertEqual(format_percent(0), "0.00%")

    def test_format_timedelta(self):
        """Test the format_timedelta function."""
        self.assertEqual(format_timedelta(3600), "1:00:00")
        self.assertEqual(format_timedelta(3661), "1:01:01")

    def test_parse_hw_sleep(self):
        """Test the parse_hw_sleep function."""
        self.assertEqual(parse_hw_sleep(0.5), 50)
        self.assertEqual(parse_hw_sleep(1.0), 100)
        self.assertEqual(parse_hw_sleep(1.5), 0)


class TestSleepReport(unittest.TestCase):
    """Unit tests for the SleepReport class."""

    @patch("amd_debug.sleep_report.SleepDatabase")
    def setUp(self, MockSleepDatabase):
        """Set up a mock SleepReport instance for testing."""
        self.mock_db = MockSleepDatabase.return_value
        self.mock_db.report_summary_dataframe.return_value = pd.DataFrame(
            {
                "t0": [datetime(2023, 10, 10, 12, 0, 0).strftime("%Y%m%d%H%M%S")],
                "t1": [datetime(2023, 10, 10, 12, 30, 0).strftime("%Y%m%d%H%M%S")],
                "hw": [50],
                "requested": [1],
                "gpio": ["1, 2"],
                "wake_irq": ["1"],
                "b0": [90],
                "b1": [85],
                "full": [100],
            }
        )
        self.since = datetime(2023, 10, 9, 0, 0, 0)
        self.until = datetime(2023, 10, 12, 0, 0, 0)
        self.report = SleepReport(
            since=self.since,
            until=self.until,
            fname=None,
            fmt="txt",
            tool_debug=False,
            report_debug=False,
        )

    def test_analyze_duration(self):
        """Test the analyze_duration method."""
        self.report.analyze_duration(
            index=0,
            t0=datetime(2023, 10, 10, 12, 0, 0),
            t1=datetime(2023, 10, 10, 12, 30, 0),
            requested=20,
            hw=50,
        )
        self.assertEqual(len(self.report.failures), 2)

    @patch("amd_debug.sleep_report.Environment")
    @patch("amd_debug.sleep_report.FileSystemLoader")
    def test_build_template(self, _mock_fsl, mock_env):
        """Test the build_template method."""
        mock_template = mock_env.return_value.get_template.return_value
        mock_template.render.return_value = "Rendered Template"
        result = self.report.build_template(inc_prereq=False)
        self.assertEqual(result, "Rendered Template")

    @patch("matplotlib.pyplot.savefig")
    def test_build_battery_chart(self, mock_savefig):
        """Test the build_battery_chart method."""
        self.report.build_battery_chart()
        self.assertIsNotNone(self.report.battery_svg)
        mock_savefig.assert_called_once()

    @patch("matplotlib.pyplot.savefig")
    def test_build_hw_sleep_chart(self, mock_savefig):
        """Test the build_hw_sleep_chart method."""
        self.report.build_hw_sleep_chart()
        self.assertIsNotNone(self.report.hwsleep_svg)
        mock_savefig.assert_called_once()

    def test_pre_process_dataframe_zero_duration(self):
        """Test the pre_process_dataframe method when t0 and t1 are the same."""
        # Mock the dataframe with t0 and t1 being the same
        self.report.df = pd.DataFrame(
            {
                "t0": [datetime(2023, 10, 10, 12, 0, 0).strftime("%Y%m%d%H%M%S")],
                "t1": [datetime(2023, 10, 10, 12, 0, 0).strftime("%Y%m%d%H%M%S")],
                "hw": [50],
                "requested": [1],
                "gpio": ["1, 2"],
                "wake_irq": ["1"],
                "b0": [90],
                "b1": [85],
                "full": [100],
            }
        )

        # Call the method
        self.report.pre_process_dataframe()

        # Verify the dataframe was processed correctly
        self.assertTrue(
            self.report.df["Duration"].isna().iloc[0]
        )  # Duration should be NaN
        self.assertTrue(
            math.isnan(self.report.df["Hardware Sleep"].iloc[0])
        )  # Hardware Sleep should be NaN



#!/usr/bin/python3
# SPDX-License-Identifier: MIT

"""
This module contains unit tests for the failure functions in the amd-debug-tools package.
"""

from unittest.mock import patch, call

import logging
import unittest
import os

import amd_debug.failures


class TestFailures(unittest.TestCase):
    """Test failure functions"""

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(filename="/dev/null", level=logging.DEBUG)

    @patch("builtins.print")
    def test_failures(self, mocked_print):
        """Test failure functions"""

        cls = amd_debug.failures.RtcAlarmWrong()
        self.assertEqual(
            cls.get_description(), "rtc_cmos is not configured to use ACPI alarm"
        )
        self.assertEqual(
            str(cls),
            "Some problems can occur during wakeup cycles if the HPET RTC "
            "emulation is used to wake systems. This can manifest in unexpected "
            "wakeups or high power consumption.For more information on this failure "
            "see:https://github.com/systemd/systemd/issues/24279",
        )
        cls = amd_debug.failures.MissingAmdgpu()
        self.assertEqual(cls.get_description(), "AMDGPU driver is missing")
        cls = amd_debug.failures.MissingAmdgpuFirmware(["foo", "bar"])
        self.assertEqual(cls.get_description(), "AMDGPU firmware is missing")
        cls = amd_debug.failures.AmdgpuPpFeatureMask()
        self.assertEqual(cls.get_description(), "AMDGPU ppfeaturemask changed")
        cls = amd_debug.failures.MissingAmdPmc()
        self.assertEqual(cls.get_description(), "AMD-PMC driver is missing")
        cls = amd_debug.failures.MissingThunderbolt()
        self.assertEqual(cls.get_description(), "thunderbolt driver is missing")
        cls = amd_debug.failures.MissingXhciHcd()
        self.assertEqual(cls.get_description(), "xhci_hcd driver is missing")
        cls = amd_debug.failures.MissingDriver("4")
        self.assertEqual(cls.get_description(), "4 driver is missing")
        cls = amd_debug.failures.AcpiBiosError("5")
        self.assertEqual(cls.get_description(), "ACPI BIOS Errors detected")
        cls = amd_debug.failures.UnsupportedModel()
        self.assertEqual(cls.get_description(), "Unsupported CPU model")
        cls = amd_debug.failures.UserNvmeConfiguration()
        self.assertEqual(cls.get_description(), "NVME ACPI support is disabled")
        cls = amd_debug.failures.AcpiNvmeStorageD3Enable("foo", 2)
        self.assertEqual(cls.get_description(), "foo missing ACPI attributes")
        cls = amd_debug.failures.DevSlpHostIssue()
        self.assertEqual(
            cls.get_description(), "AHCI controller doesn't support DevSlp"
        )
        cls = amd_debug.failures.DevSlpDiskIssue()
        self.assertEqual(cls.get_description(), "SATA disk doesn't support DevSlp")
        cls = amd_debug.failures.SleepModeWrong()
        self.assertEqual(
            cls.get_description(),
            "The system hasn't been configured for Modern Standby in BIOS setup",
        )
        cls = amd_debug.failures.DeepSleep()
        self.assertEqual(
            cls.get_description(),
            "The kernel command line is asserting the system to use deep sleep",
        )
        cls = amd_debug.failures.FadtWrong()
        self.assertEqual(
            cls.get_description(),
            "The kernel didn't emit a message that low power idle was supported",
        )
        cls = amd_debug.failures.Irq1Workaround()
        self.assertEqual(
            cls.get_description(),
            "The wakeup showed an IRQ1 wakeup source, which might be a platform firmware bug",
        )
        cls = amd_debug.failures.KernelRingBufferWrapped()
        self.assertEqual(cls.get_description(), "Kernel ringbuffer has wrapped")
        cls = amd_debug.failures.AmdHsmpBug()
        self.assertEqual(cls.get_description(), "amd-hsmp built in to kernel")
        cls = amd_debug.failures.WCN6855Bug()
        self.assertEqual(
            cls.get_description(),
            "The firmware loaded for the WCN6855 causes spurious wakeups",
        )
        cls = amd_debug.failures.I2CHidBug("touchpad", "block")
        self.assertEqual(
            cls.get_description(),
            "The touchpad device has been reported to cause high power consumption and spurious wakeups",
        )
        cls = amd_debug.failures.SpuriousWakeup(1, 0)
        self.assertEqual(
            cls.get_description(), "Userspace wasn't asleep at least 0:00:01"
        )
        cls = amd_debug.failures.LowHardwareSleepResidency(5, 30)
        self.assertEqual(
            cls.get_description(), "System had low hardware sleep residency"
        )
        cls = amd_debug.failures.MSRFailure()
        self.assertEqual(cls.get_description(), "PC6 or CC6 state disabled")
        cls = amd_debug.failures.TaintedKernel()
        self.assertEqual(cls.get_description(), "Kernel is tainted")
        cls = amd_debug.failures.DMArNotEnabled()
        self.assertEqual(cls.get_description(), "Pre-boot DMA protection disabled")
        cls = amd_debug.failures.MissingIommuACPI("foo")
        self.assertEqual(cls.get_description(), "Device foo missing from ACPI tables")
        cls = amd_debug.failures.MissingIommuPolicy("foo")
        self.assertEqual(
            cls.get_description(), "Device foo does not have IOMMU policy applied"
        )
        cls = amd_debug.failures.IommuPageFault("foo")
        self.assertEqual(cls.get_description(), "Page fault reported for foo")
        cls = amd_debug.failures.SMTNotEnabled()
        self.assertEqual(cls.get_description(), "SMT is not enabled")
        cls = amd_debug.failures.ASpmWrong()
        self.assertEqual(cls.get_description(), "ASPM is overridden")
        cls = amd_debug.failures.UnservicedGpio()
        self.assertEqual(cls.get_description(), "GPIO interrupt is not serviced")
        cls = amd_debug.failures.DmiNotSetup()
        self.assertEqual(cls.get_description(), "DMI data was not scanned")
        cls = amd_debug.failures.LimitedCores(10, 7)
        self.assertEqual(cls.get_description(), "CPU cores have been limited")
        cls = amd_debug.failures.RogAllyOldMcu(1, 2)
        self.assertEqual(cls.get_description(), "Rog Ally MCU firmware is too old")
        os.environ["TERM"] = "dumb"
        cls = amd_debug.failures.RogAllyMcuPowerSave()
        self.assertEqual(cls.get_description(), "Rog Ally MCU power save is disabled")
        failure = "The MCU powersave feature is disabled which will cause problems with the controller after suspend/resume."
        self.assertEqual(str(cls), failure)
        cls.get_failure()
        mocked_print.assert_has_calls(
            [
                call("🚦 Rog Ally MCU power save is disabled"),
                call(failure),
            ]
        )



#!/usr/bin/python3
# SPDX-License-Identifier: MIT

"""
This module contains unit tests for the datbase functions in the amd-debug-tools package.
"""
import unittest

from datetime import datetime

from amd_debug.database import SleepDatabase


class TestSleepDatabase(unittest.TestCase):
    """Test SleepDatabase class"""

    def setUp(self):
        """Set up mocks and an in-memory database for testing"""
        # Initialize SleepDatabase after mocks are set up
        self.db = SleepDatabase(dbf=":memory:")

    def test_start_cycle(self):
        """Test starting a new sleep cycle"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.assertEqual(self.db.last_suspend, timestamp)
        self.assertEqual(self.db.cycle_data_cnt, 0)
        self.assertEqual(self.db.debug_cnt, 0)

    def test_record_debug(self):
        """Test recording a debug message"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_debug("Test debug message", level=5)
        cur = self.db.db.cursor()
        cur.execute(
            "SELECT message, priority FROM debug WHERE t0=?",
            (int(timestamp.strftime("%Y%m%d%H%M%S")),),
        )
        result = cur.fetchone()
        self.assertEqual(result, ("Test debug message", 5))

    def test_record_battery_energy(self):
        """Test recording battery energy"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_battery_energy("Battery1", 50, 100, "mWh")
        cur = self.db.db.cursor()
        cur.execute(
            "SELECT name, b0, b1, full, unit FROM battery WHERE t0=?",
            (int(timestamp.strftime("%Y%m%d%H%M%S")),),
        )
        result = cur.fetchone()
        self.assertEqual(result, ("Battery1", 50, None, 100, "mWh"))

    def test_record_cycle_data(self):
        """Test recording cycle data"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_cycle_data("Test cycle data", "symbol1")
        cur = self.db.db.cursor()
        cur.execute(
            "SELECT message, symbol FROM cycle_data WHERE t0=?",
            (int(timestamp.strftime("%Y%m%d%H%M%S")),),
        )
        result = cur.fetchone()
        self.assertEqual(result, ("Test cycle data", "symbol1"))

    def test_record_cycle(self):
        """Test recording a sleep cycle"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_cycle(
            requested_duration=100,
            active_gpios="GPIO1",
            wakeup_irqs="IRQ1",
            kernel_duration=1.5,
            hw_sleep_duration=2.5,
        )
        cur = self.db.db.cursor()
        cur.execute(
            "SELECT requested, gpio, wake_irq, kernel, hw FROM cycle WHERE t0=?",
            (int(timestamp.strftime("%Y%m%d%H%M%S")),),
        )
        result = cur.fetchone()
        self.assertEqual(result, (100, "GPIO1", "IRQ1", 1.5, 2.5))

    def test_report_debug(self):
        """Test reporting debug messages"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_debug("Test debug message", level=5)
        result = self.db.report_debug(timestamp)
        self.assertEqual(result, [("Test debug message", 5)])

    def test_report_battery(self):
        """Test reporting battery data"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_battery_energy("Battery1", 50, 100, "mWh")
        result = self.db.report_battery(timestamp)
        self.assertEqual(
            result,
            [
                (
                    int(timestamp.strftime("%Y%m%d%H%M%S")),
                    "Battery1",
                    50,
                    None,
                    100,
                    "mWh",
                )
            ],
        )

    def test_record_prereq(self):
        """Test recording a prereq message"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_prereq("Test prereq message", "symbol1")
        cur = self.db.db.cursor()
        cur.execute(
            "SELECT message, symbol FROM prereq_data WHERE t0=?",
            (int(timestamp.strftime("%Y%m%d%H%M%S")),),
        )
        result = cur.fetchone()
        self.assertEqual(result, ("Test prereq message", "symbol1"))

    def test_report_prereq(self):
        """Test reporting prereq messages"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_prereq("Test prereq message 1", "symbol1")
        self.db.record_prereq("Test prereq message 2", "symbol2")
        result = self.db.report_prereq(timestamp)
        self.assertEqual(
            result,
            [
                (
                    int(timestamp.strftime("%Y%m%d%H%M%S")),
                    0,
                    "Test prereq message 1",
                    "symbol1",
                ),
                (
                    int(timestamp.strftime("%Y%m%d%H%M%S")),
                    1,
                    "Test prereq message 2",
                    "symbol2",
                ),
            ],
        )

    def test_report_prereq_no_data(self):
        """Test reporting prereq messages when no data exists"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        result = self.db.report_prereq(timestamp)
        self.assertEqual(result, [])

    def test_report_prereq_none_timestamp(self):
        """Test reporting prereq messages with None timestamp"""
        result = self.db.report_prereq(None)
        self.assertEqual(result, [])

    def test_report_cycle(self):
        """Test reporting a cycle from the database"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_cycle(
            requested_duration=100,
            active_gpios="GPIO1",
            wakeup_irqs="IRQ1",
            kernel_duration=1.5,
            hw_sleep_duration=2.5,
        )
        result = self.db.report_cycle(timestamp)
        self.assertEqual(
            result,
            [
                (
                    int(timestamp.strftime("%Y%m%d%H%M%S")),
                    int(datetime.now().strftime("%Y%m%d%H%M%S")),
                    100,
                    "GPIO1",
                    "IRQ1",
                    1.5,
                    2.5,
                )
            ],
        )

    def test_report_cycle_no_data(self):
        """Test reporting a cycle when no data exists"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        result = self.db.report_cycle(timestamp)
        self.assertEqual(result, [])

    def test_report_cycle_none_timestamp(self):
        """Test reporting a cycle with None timestamp"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_cycle(
            requested_duration=100,
            active_gpios="GPIO1",
            wakeup_irqs="IRQ1",
            kernel_duration=1.5,
            hw_sleep_duration=2.5,
        )
        result = self.db.report_cycle(None)
        self.assertEqual(
            result,
            [
                (
                    int(timestamp.strftime("%Y%m%d%H%M%S")),
                    int(datetime.now().strftime("%Y%m%d%H%M%S")),
                    100,
                    "GPIO1",
                    "IRQ1",
                    1.5,
                    2.5,
                )
            ],
        )

    def test_get_last_cycle(self):
        """Test getting the last cycle from the database"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_cycle(
            requested_duration=100,
            active_gpios="GPIO1",
            wakeup_irqs="IRQ1",
            kernel_duration=1.5,
            hw_sleep_duration=2.5,
        )
        result = self.db.get_last_cycle()
        self.assertEqual(result, (int(timestamp.strftime("%Y%m%d%H%M%S")),))

    def test_get_last_cycle_no_data(self):
        """Test getting the last cycle when no data exists"""
        result = self.db.get_last_cycle()
        self.assertIsNone(result)

    def test_get_last_prereq_ts(self):
        """Test getting the last prereq timestamp from the database"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_prereq("Test prereq message 1", "symbol1")
        self.db.record_prereq("Test prereq message 2", "symbol2")
        result = self.db.get_last_prereq_ts()
        self.assertEqual(result, int(timestamp.strftime("%Y%m%d%H%M%S")))

    def test_get_last_prereq_ts_no_data(self):
        """Test getting the last prereq timestamp when no data exists"""
        result = self.db.get_last_prereq_ts()
        self.assertEqual(result, 0)

    def test_report_cycle_data(self):
        """Test reporting cycle data"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_cycle_data("Test cycle data 1", "symbol1")
        self.db.record_cycle_data("Test cycle data 2", "symbol2")
        result = self.db.report_cycle_data(timestamp)
        expected_result = "symbol1 Test cycle data 1\nsymbol2 Test cycle data 2\n"
        self.assertEqual(result, expected_result)

    def test_report_cycle_data_no_data(self):
        """Test reporting cycle data when no data exists"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        result = self.db.report_cycle_data(timestamp)
        self.assertEqual(result, "")

    def test_report_cycle_data_none_timestamp(self):
        """Test reporting cycle data with None timestamp"""
        timestamp = datetime.now()
        self.db.start_cycle(timestamp)
        self.db.record_cycle_data("Test cycle data 1", "symbol1")
        result = self.db.report_cycle_data(None)
        expected_result = "symbol1 Test cycle data 1\n"
        self.assertEqual(result, expected_result)



# SPDX-License-Identifier: MIT

"""
This module contains installer support for amd-debug-tools.
"""

import argparse
import os
import shutil
import subprocess
from amd_debug.common import (
    print_color,
    get_distro,
    read_file,
    systemd_in_use,
    show_log_info,
    fatal_error,
    relaunch_sudo,
    AmdTool,
)


class Headers:  # pylint: disable=too-few-public-methods
    """Headers for the script"""

    MissingIasl = "ACPI extraction tool `iasl` is missing"
    MissingEdidDecode = "EDID decoding tool `edid-decode` is missing"
    MissingDiEdidDecode = "EDID decoding tool `di-edid-decode` is missing"
    MissingEthtool = "Ethtool is missing"
    InstallAction = "Attempting to install"
    MissingFwupd = "Firmware update library `fwupd` is missing"
    MissingPyudev = "Udev access library `pyudev` is missing"
    MissingPackaging = "Python library `packaging` is missing"
    MissingPandas = "Data library `pandas` is missing"
    MissingTabulate = "Data library `tabulate` is missing"
    MissingJinja2 = "Template library `jinja2` is missing"
    MissingSeaborn = "Data visualization library `seaborn` is missing"
    UnknownDistro = "No distro installation support available, install manually"


class DistroPackage:
    """Base class for distro packages"""

    def __init__(self, deb, rpm, arch, message):
        self.deb = deb
        self.rpm = rpm
        self.arch = arch
        self.message = message

    def install(self):
        """Install the package for a given distro"""
        relaunch_sudo()
        show_install_message(self.message)
        dist = get_distro()
        if dist in ("ubuntu", "debian"):
            if not self.deb:
                return False
            installer = ["apt", "install", self.deb]
        elif dist == "fedora":
            if not self.rpm:
                return False
            release = read_file("/usr/lib/os-release")
            variant = None
            for line in release.split("\n"):
                if line.startswith("VARIANT_ID"):
                    variant = line.split("=")[-1]
            if variant not in ("workstation", "kde"):
                return False
            installer = ["dnf", "install", "-y", self.rpm]
        elif dist == "arch" or os.path.exists("/etc/arch-release"):
            if not self.arch:
                return False
            installer = ["pacman", "-Sy", self.arch]
        else:
            print_color(Headers.UnknownDistro, "👀")
            return True

        try:
            subprocess.check_call(installer)
        except subprocess.CalledProcessError as e:
            fatal_error(e)
        return True


class PyUdevPackage(DistroPackage):
    """Pyudev package"""

    def __init__(self):
        super().__init__(
            deb="python3-pyudev",
            rpm="python3-pyudev",
            arch="python-pyudev",
            message=Headers.MissingPyudev,
        )


class PackagingPackage(DistroPackage):
    """Packaging package"""

    def __init__(self):
        super().__init__(
            deb="python3-packaging",
            rpm=None,
            arch="python-packaging",
            message=Headers.MissingPackaging,
        )


class PandasPackage(DistroPackage):
    """Class for handling the pandas package"""

    def __init__(self):
        super().__init__(
            deb="python3-pandas",
            rpm="python3-pandas",
            arch="python-pandas",
            message=Headers.MissingPandas,
        )


class TabulatePackage(DistroPackage):
    """Class for handling the tabulate package"""

    def __init__(self):
        super().__init__(
            deb="python3-tabulate",
            rpm="python3-tabulate",
            arch="python-tabulate",
            message=Headers.MissingTabulate,
        )


class Jinja2Package(DistroPackage):
    """Class for handling the jinja2 package"""

    def __init__(self):
        super().__init__(
            deb="python3-jinja2",
            rpm="python3-jinja2",
            arch="python-jinja",
            message=Headers.MissingJinja2,
        )


class SeabornPackage(DistroPackage):
    """Class for handling the seaborn package"""

    def __init__(self):
        super().__init__(
            deb="python3-seaborn",
            rpm="python3-seaborn",
            arch="python-seaborn",
            message=Headers.MissingSeaborn,
        )


class IaslPackage(DistroPackage):
    """Iasl package"""

    def __init__(self):
        super().__init__(
            deb="acpica-tools",
            rpm="acpica-tools",
            arch="acpica",
            message=Headers.MissingIasl,
        )


class EthtoolPackage(DistroPackage):
    """Ethtool package"""

    def __init__(self):
        super().__init__(
            deb="ethtool",
            rpm="ethtool",
            arch="ethtool",
            message=Headers.MissingEthtool,
        )


class EdidDecodePackage(DistroPackage):
    """Edid-Decode package"""

    def __init__(self):
        super().__init__(
            deb="edid-decode",
            rpm="edid-decode",
            arch=None,
            message=Headers.MissingEdidDecode,
        )


class DisplayInfoPackage(DistroPackage):
    """display info package"""

    def __init__(self):
        super().__init__(
            deb="libdisplay-info-bin",
            rpm="libdisplay-info-tools",
            arch="libdisplay-info",
            message=Headers.MissingDiEdidDecode,
        )


class FwupdPackage(DistroPackage):
    """Fwupd package"""

    def __init__(self):
        super().__init__(
            deb="gir1.2-fwupd-2.0",
            rpm=None,
            arch=None,
            message=Headers.MissingFwupd,
        )


def show_install_message(message):
    """Show an install message"""
    action = Headers.InstallAction
    message = f"{message}. {action}."
    print_color(message, "👀")


class Installer(AmdTool):
    """Installer class"""

    def __init__(self, tool_debug):
        log_prefix = "installer" if tool_debug else None
        super().__init__(log_prefix)
        self.systemd = systemd_in_use()
        self.systemd_path = os.path.join("/", "lib", "systemd", "system-sleep")

        # test if fwupd can report device firmware versions
        try:
            import gi  # pylint: disable=import-outside-toplevel
            from gi.repository import (  # pylint: disable=import-outside-toplevel
                GLib as _,
            )

            gi.require_version("Fwupd", "2.0")
            from gi.repository import (  # pylint: disable=import-outside-toplevel
                Fwupd as _,
            )

            self.fwupd = True
        except ImportError:
            self.fwupd = False
        except ValueError:
            self.fwupd = False
        self.requirements = []

    def set_requirements(self, *args):
        """Set the requirements for the installer"""
        self.requirements = args

    def install_dependencies(self) -> bool:
        """Install the dependencies"""
        if "iasl" in self.requirements:
            try:
                iasl = subprocess.call(["iasl", "-v"], stdout=subprocess.DEVNULL) == 0
            except FileNotFoundError:
                iasl = False
            if not iasl:
                package = IaslPackage()
                if not package.install():
                    return False
        if "ethtool" in self.requirements:
            try:
                ethtool = (
                    subprocess.call(["ethtool", "-h"], stdout=subprocess.DEVNULL) == 0
                )
            except FileNotFoundError:
                ethtool = False
            if not ethtool:
                package = EthtoolPackage()
                if not package.install():
                    return False
        # can be satisified by either edid-decode or di-edid-decode
        if "edid-decode" in self.requirements:
            try:
                di_edid = (
                    subprocess.call(
                        ["di-edid-decode", "--help"],
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )
                    == 255
                )
            except FileNotFoundError:
                di_edid = False
            try:
                edid = (
                    subprocess.call(
                        ["edid-decode", "--help"], stdout=subprocess.DEVNULL
                    )
                    == 255
                )
            except FileNotFoundError:
                edid = False
            if not di_edid and not edid:
                # try to install di-edid-decode first
                package = DisplayInfoPackage()
                if package.install():
                    return True
                # fall back to edid-decode instead
                package = EdidDecodePackage()
                if not package.install():
                    return False
        if "fwupd" in self.requirements and not self.fwupd:
            package = FwupdPackage()
            if not package.install():
                return False
        if "pyudev" in self.requirements:
            try:
                import pyudev as _  # pylint: disable=import-outside-toplevel
            except ModuleNotFoundError:
                package = PyUdevPackage()
                if not package.install():
                    return False
        if "packaging" in self.requirements:
            try:
                import packaging as _  # pylint: disable=import-outside-toplevel
            except ModuleNotFoundError:
                package = PackagingPackage()
                if not package.install():
                    return False
        if "pandas" in self.requirements:
            try:
                import pandas as _  # pylint: disable=import-outside-toplevel
            except ModuleNotFoundError:
                package = PandasPackage()
                if not package.install():
                    return False
        if "tabulate" in self.requirements:
            try:
                import tabulate as _  # pylint: disable=import-outside-toplevel
            except ModuleNotFoundError:
                package = TabulatePackage()
                if not package.install():
                    return False
        if "jinja2" in self.requirements:
            try:
                import jinja2 as _  # pylint: disable=import-outside-toplevel
            except ModuleNotFoundError:
                package = Jinja2Package()
                if not package.install():
                    return False
        if "seaborn" in self.requirements:
            try:
                import seaborn as _  # pylint: disable=import-outside-toplevel
            except ModuleNotFoundError:
                package = SeabornPackage()
                if not package.install():
                    return False

        return True

    def _check_systemd(self) -> bool:
        """Check if the systemd path exists"""
        if not os.path.exists(self.systemd_path):
            print_color(
                f"Systemd path does not exist: {self.systemd_path}",
                "❌",
            )
        return os.path.exists(self.systemd_path)

    def remove(self) -> bool:
        """Remove the amd-s2idle hook"""
        if self._check_systemd():
            f = "s2idle-hook"
            t = os.path.join(self.systemd_path, f)
            os.remove(t)
            print_color(
                f"Removed {f} from {self.systemd_path}",
                "✅",
            )
        else:
            print_color("Systemd path does not exist, not removing hook", "🚦")
        f = "amd-s2idle"
        d = os.path.join(
            "/",
            "usr",
            "local",
            "share",
            "bash-completion",
            "completions",
        )
        t = os.path.join(d, f)
        if os.path.exists(t):
            os.remove(t)
            print_color(f"Removed {f} from {d}", "✅")
        return True

    def install(self) -> bool:
        """Install the amd-s2idle hook"""
        import amd_debug  # pylint: disable=import-outside-toplevel

        d = os.path.dirname(amd_debug.__file__)
        if self._check_systemd():
            f = "s2idle-hook"
            s = os.path.join(d, f)
            t = os.path.join(self.systemd_path, f)
            with open(s, "r", encoding="utf-8") as r:
                with open(t, "w", encoding="utf-8") as w:
                    for line in r.readlines():
                        if 'parser.add_argument("--path"' in line:
                            line = line.replace(
                                'default=""',
                                f"default=\"{os.path.join(d, '..')}\"",
                            )
                        w.write(line)
            os.chmod(t, 0o755)
            print_color(
                f"Installed {f} to {self.systemd_path}",
                "✅",
            )
        else:
            print_color("Systemd path does not exist, not installing hook", "🚦")
        f = "amd-s2idle"
        s = os.path.join(d, "bash", f)
        t = os.path.join(
            "/",
            "usr",
            "local",
            "share",
            "bash-completion",
            "completions",
        )
        os.makedirs(t, exist_ok=True)
        shutil.copy(s, t)
        print_color(f"Installed {f} to {t}", "✅")
        return True


def parse_args():
    """Parse command line arguments"""
    parser = argparse.ArgumentParser(
        description="Install dependencies for AMD debug tools",
    )
    parser.add_argument(
        "--tool-debug",
        action="store_true",
        help="Enable tool debug logging",
    )
    return parser.parse_args()


def install_dep_superset() -> None | int:
    """Install all python superset dependencies"""
    args = parse_args()
    tool = Installer(tool_debug=args.tool_debug)
    tool.set_requirements(
        "iasl",
        "ethtool",
        "jinja2",
        "pyudev",
        "packaging",
        "pandas",
        "seaborn",
        "tabulate",
        "edid-decode",
    )
    ret = tool.install_dependencies()
    if ret:
        print_color("All dependencies installed", "✅")
    show_log_info()
    if ret is False:
        return 1
    return



#!/usr/bin/python3
# SPDX-License-Identifier: MIT

"""
This module contains unit tests for the display functions in the amd-debug-tools package.
"""

import logging
import unittest
from unittest.mock import patch, MagicMock
from amd_debug.display import Display


class TestDisplay(unittest.TestCase):
    """Unit tests for the Display class in amd-debug-tools"""

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(filename="/dev/null", level=logging.DEBUG)

    @patch("amd_debug.display.Context")
    @patch("amd_debug.display.read_file")
    @patch("os.path.exists")
    def test_display_initialization(self, mock_exists, mock_read_file, mock_context):
        """Test the Display class initialization and EDID retrieval"""
        # Mock the pyudev Context and its list_devices method
        mock_device = MagicMock()
        mock_device.device_path = "/devices/card0"
        mock_device.sys_path = "/sys/devices/card0"
        mock_device.sys_name = "card0"
        mock_context.return_value.list_devices.return_value = [mock_device]

        # Mock os.path.exists and read_file behavior
        mock_exists.side_effect = lambda path: "status" in path or "enabled" in path
        mock_read_file.side_effect = lambda path: (
            "connected" if "status" in path else "enabled"
        )

        # Initialize the Display class
        display = Display()

        # Verify the EDID paths are correctly set
        expected_edid = ["/sys/devices/card0/edid"]
        self.assertEqual(display.get_edid(), expected_edid)
        mock_context.assert_called_once()

    @patch("amd_debug.display.Context")
    def test_no_connected_displays(self, mock_context):
        """Test the Display class when no connected displays are found"""
        # Mock the pyudev Context to return no devices
        mock_context.return_value.list_devices.return_value = []

        # Initialize the Display class
        display = Display()

        # Verify the EDID dictionary is empty
        self.assertEqual(display.get_edid(), [])

    @patch("amd_debug.display.Context")
    def test_device_without_card(self, mock_context):
        """Test the Display class with a device that does not have 'card' in the name"""
        # Mock the pyudev Context to return a device without 'card' in the name
        mock_device = MagicMock()
        mock_device.device_path = "/devices/other_device"
        mock_device.sys_path = "/sys/devices/other_device"
        mock_device.sys_name = "other_device"
        mock_context.return_value.list_devices.return_value = [mock_device]

        # Initialize the Display class
        display = Display()

        # Verify the EDID dictionary is empty
        self.assertEqual(display.get_edid(), [])

    @patch("amd_debug.display.Context")
    @patch("amd_debug.display.read_file")
    @patch("os.path.exists")
    def test_device_not_enabled(self, mock_exists, mock_read_file, mock_context):
        """Test the Display class with a device that is not enabled"""
        # Mock the pyudev Context to return a device that is not enabled
        mock_device = MagicMock()
        mock_device.device_path = "/devices/card0"
        mock_device.sys_path = "/sys/devices/card0"
        mock_device.sys_name = "card0"
        mock_context.return_value.list_devices.return_value = [mock_device]

        # Mock os.path.exists and read_file behavior
        mock_exists.side_effect = lambda path: "status" in path or "enabled" in path
        mock_read_file.side_effect = lambda path: (
            "connected" if "status" in path else "disabled"
        )

        # Initialize the Display class
        display = Display()

        # Verify the EDID dictionary is empty
        self.assertEqual(display.get_edid(), [])

    @patch("amd_debug.display.Context")
    @patch("amd_debug.display.read_file")
    @patch("os.path.exists")
    def test_missing_status_file(self, mock_exists, mock_read_file, mock_context):
        """Test the Display class when the status file is missing"""
        # Mock the pyudev Context to return a device with a missing status file
        mock_device = MagicMock()
        mock_device.device_path = "/devices/card0"
        mock_device.sys_path = "/sys/devices/card0"
        mock_device.sys_name = "card0"
        mock_context.return_value.list_devices.return_value = [mock_device]

        # Mock os.path.exists to return False for the status file
        mock_exists.side_effect = lambda path: "enabled" in path
        mock_read_file.side_effect = lambda path: "enabled" if "enabled" in path else ""

        # Initialize the Display class
        display = Display()

        # Verify the EDID dictionary is empty
        self.assertEqual(display.get_edid(), [])

    @patch("amd_debug.display.Context")
    @patch("amd_debug.display.read_file")
    @patch("os.path.exists")
    def test_status_not_connected(self, mock_exists, mock_read_file, mock_context):
        """Test the Display class when the status file does not indicate connected"""
        # Mock the pyudev Context to return a device with a status file that does not indicate connected
        mock_device = MagicMock()
        mock_device.device_path = "/devices/card0"
        mock_device.sys_path = "/sys/devices/card0"
        mock_device.sys_name = "card0"
        mock_context.return_value.list_devices.return_value = [mock_device]

        # Mock os.path.exists and read_file behavior
        mock_exists.side_effect = lambda path: "status" in path or "enabled" in path
        mock_read_file.side_effect = lambda path: (
            "not_connected" if "status" in path else "enabled"
        )

        # Initialize the Display class
        display = Display()

        # Verify the EDID dictionary is empty
        self.assertEqual(display.get_edid(), [])



# SPDX-License-Identifier: MIT
"""s2idle analysis tool"""
import argparse
import re
import sys
from amd_debug.common import (
    AmdTool,
    fatal_error,
    get_log_priority,
    minimum_kernel,
    print_color,
    relaunch_sudo,
    show_log_info,
    version,
)
from amd_debug.kernel import get_kernel_log, sscanf_bios_args
from amd_debug.acpi import AcpicaTracer


class AmdBios(AmdTool):
    """
    AmdBios is a class which fetches the BIOS events from kernel logs.
    """

    def __init__(self, inf, debug):
        log_prefix = "bios" if debug else None
        super().__init__(log_prefix)
        self.kernel_log = get_kernel_log(inf)

    def set_tracing(self, enable):
        """Run the action"""
        relaunch_sudo()

        if not minimum_kernel(6, 16):
            print_color(
                "Support for BIOS debug logging was merged in mainline 6.16, "
                "this tool may not work correctly unless support is manually "
                "backported",
                "🚦",
            )

        tracer = AcpicaTracer()
        ret = tracer.trace_bios() if enable else tracer.disable()
        if ret:
            action = "enabled" if enable else "disabled"
            print_color(f"Set BIOS tracing to {action}", "✅")
        else:
            fatal_error(
                "BIOS tracing not supported, please check your kernel for CONFIG_ACPI_DEBUG"
            )

        return True

    def _analyze_kernel_log_line(self, line, priority):
        """Analyze a line from the kernel log"""
        bios_args = sscanf_bios_args(line)
        if bios_args:
            if isinstance(bios_args, str):
                print_color(bios_args, "🖴")
            else:
                return
        else:
            # strip timestamp
            t = re.sub(r"^\[\s*\d+\.\d+\]", "", line).strip()
            print_color(t, get_log_priority(priority))

    def run(self):
        """Exfiltrate from the kernel log"""
        self.kernel_log.process_callback(self._analyze_kernel_log_line)
        return True


def parse_args():
    """Parse command line arguments"""
    parser = argparse.ArgumentParser(
        description="Parse a combined kernel/BIOS log.",
    )
    subparsers = parser.add_subparsers(help="Possible commands", dest="command")
    parse_cmd = subparsers.add_parser(
        "parse", help="Parse log for kernel and BIOS messages"
    )
    parse_cmd.add_argument(
        "--input",
        help="Optional input file to parse",
    )
    parse_cmd.add_argument(
        "--tool-debug",
        action="store_true",
        help="Enable tool debug logging",
    )
    trace_cmd = subparsers.add_parser("trace", help="Enable or disable tracing")
    trace_cmd.add_argument(
        "--enable",
        action="store_true",
        help="Enable BIOS AML tracing",
    )
    trace_cmd.add_argument(
        "--disable",
        action="store_true",
        help="Disable BIOS AML tracing",
    )
    trace_cmd.add_argument(
        "--tool-debug",
        action="store_true",
        help="Enable tool debug logging",
    )
    parser.add_argument(
        "--version", action="store_true", help="Show version information"
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    if args.command == "trace":
        if args.enable and args.disable:
            sys.exit("can't set both enable and disable")
        if not args.enable and not args.disable:
            sys.exit("must set either enable or disable")

    return args


def main() -> None | int:
    """Main function"""
    args = parse_args()
    ret = False
    if args.command == "trace":
        app = AmdBios(None, args.tool_debug)
        ret = app.set_tracing(True if args.enable else False)
    elif args.command == "parse":
        app = AmdBios(args.input, args.tool_debug)
        ret = app.run()
    elif args.version:
        print(version())
    show_log_info()
    if ret is False:
        return 1
    return



# SPDX-License-Identifier: MIT
"""CPPC triage script for AMD systems"""

import os
import argparse
import re
import sys

import pandas as pd
from tabulate import tabulate
from pyudev import Context

from amd_debug.common import (
    AmdTool,
    get_pretty_distro,
    print_color,
    read_file,
    read_msr,
    relaunch_sudo,
    show_log_info,
    version,
)


class MSR:  # pylint: disable=too-few-public-methods
    """MSR addresses for CPPC"""

    MSR_AMD_CPPC_CAP1 = 0xC00102B0
    MSR_AMD_CPPC_ENABLE = 0xC00102B1
    MSR_AMD_CPPC_CAP2 = 0xC00102B2
    MSR_AMD_CPPC_REQ = 0xC00102B3
    MSR_AMD_CPPC_STATUS = 0xC00102B4


def amd_cppc_cap_lowest_perf(x):
    """Return the lowest performance value from the given input."""
    return x & 0xFF


def amd_cppc_cap_lownonlin_perf(x):
    """Return the lowest nonlinear performance value from the given input."""
    return (x >> 8) & 0xFF


def amd_cppc_cap_nominal_perf(x):
    """Return the nominal performance value from the given input."""
    return (x >> 16) & 0xFF


def amd_cppc_cap_highest_perf(x):
    """Return the highest performance value from the given input."""
    return (x >> 24) & 0xFF


def amd_cppc_max_perf(x):
    """Return the maximum performance value from the given input."""
    return x & 0xFF


def amd_cppc_min_perf(x):
    """Return the minimum performance value from the given input."""
    return (x >> 8) & 0xFF


def amd_cppc_des_perf(x):
    """Return the desired performance value from the given input."""
    return (x >> 16) & 0xFF


def amd_cppc_epp_perf(x):
    """Return the energy performance preference value from the given input."""
    return (x >> 24) & 0xFF


class AmdPstateTriage(AmdTool):
    """Class for handling the triage process"""

    def __init__(self, logging):
        log_prefix = "pstate" if logging else None
        super().__init__(log_prefix)
        relaunch_sudo()

        pretty = get_pretty_distro()
        print_color(f"{pretty}", "🐧")

        self.context = Context()

    def gather_amd_pstate_info(self):
        """Gather AMD Pstate global information"""
        for f in ("status", "prefcore"):
            p = os.path.join("/", "sys", "devices", "system", "cpu", "amd_pstate", f)
            if os.path.exists(p):
                print_color(f"'{f}':\t{read_file(p)}", "○")

    def gather_kernel_info(self):
        """Gather kernel information"""
        print_color(f"Kernel:\t{os.uname().release}", "🐧")

    def gather_scheduler_info(self):
        """Gather information about the scheduler"""
        procfs = os.path.join("/", "proc", "sys", "kernel", "sched_itmt_enabled")
        debugfs = os.path.join(
            "/", "sys", "kernel", "debug", "x86", "sched_itmt_enabled"
        )
        for p in [procfs, debugfs]:
            if os.path.exists(p):
                val = read_file(p)
                print_color(f"ITMT:\t{val}", "🐧")

    def gather_cpu_info(self):
        """Gather a dataframe of CPU information"""

        df = pd.DataFrame(
            columns=[
                "CPU #",
                "CPU Min Freq",
                "CPU Nonlinear Freq",
                "CPU Max Freq",
                "Scaling Min Freq",
                "Scaling Max Freq",
                "Energy Performance Preference",
                "Prefcore",
                "Boost",
            ]
        )

        for device in self.context.list_devices(subsystem="cpu"):
            p = os.path.join(device.sys_path, "cpufreq")
            if not os.path.exists(p):
                continue
            row = [
                int(re.findall(r"\d+", f"{device.sys_name}")[0]),
                read_file(os.path.join(p, "cpuinfo_min_freq")),
                read_file(os.path.join(p, "amd_pstate_lowest_nonlinear_freq")),
                read_file(os.path.join(p, "cpuinfo_max_freq")),
                read_file(os.path.join(p, "scaling_min_freq")),
                read_file(os.path.join(p, "scaling_max_freq")),
                read_file(os.path.join(p, "energy_performance_preference")),
                read_file(os.path.join(p, "amd_pstate_prefcore_ranking")),
                read_file(os.path.join(p, "boost")),
            ]
            df = pd.concat(
                [pd.DataFrame([row], columns=df.columns), df], ignore_index=True
            )

        cpuinfo = read_file("/proc/cpuinfo")
        model = re.findall(r"model name\s+:\s+(.*)", cpuinfo)[0]
        print_color(f"CPU:\t\t{model}", "💻")

        df = df.sort_values(by="CPU #")
        print_color(
            "Per-CPU sysfs files\n%s"
            % tabulate(df, headers="keys", tablefmt="psql", showindex=False),
            "🔋",
        )

    def gather_msrs(self):
        """Gather MSR information"""
        cpus = []
        for device in self.context.list_devices(subsystem="cpu"):
            cpu = int(re.findall(r"\d+", f"{device.sys_name}")[0])
            cpus.append(cpu)
        cpus.sort()

        df = pd.DataFrame(
            columns=[
                "CPU #",
                "Min Perf",
                "Max Perf",
                "Desired Perf",
                "Energy Performance Perf",
            ]
        )

        msr_df = pd.DataFrame(
            columns=[
                "CPU #",
                "Enable",
                "Status",
                "Cap 1",
                "Cap 2",
                "Request",
            ]
        )

        cap_df = pd.DataFrame(
            columns=[
                "CPU #",
                "Lowest Perf",
                "Nonlinear Perf",
                "Nominal Perf",
                "Highest Perf",
            ]
        )

        try:
            for cpu in cpus:
                enable = read_msr(MSR.MSR_AMD_CPPC_ENABLE, cpu)
                status = read_msr(MSR.MSR_AMD_CPPC_STATUS, cpu)
                cap1 = read_msr(MSR.MSR_AMD_CPPC_CAP1, cpu)
                cap2 = read_msr(MSR.MSR_AMD_CPPC_CAP2, cpu)

                req = read_msr(MSR.MSR_AMD_CPPC_REQ, cpu)
                row = [
                    cpu,
                    amd_cppc_min_perf(req),
                    amd_cppc_max_perf(req),
                    amd_cppc_des_perf(req),
                    amd_cppc_epp_perf(req),
                ]
                df = pd.concat(
                    [pd.DataFrame([row], columns=df.columns), df], ignore_index=True
                )

                row = [
                    cpu,
                    enable,
                    status,
                    hex(cap1),
                    hex(cap2),
                    hex(req),
                ]
                msr_df = pd.concat(
                    [pd.DataFrame([row], columns=msr_df.columns), msr_df],
                    ignore_index=True,
                )

                row = [
                    cpu,
                    amd_cppc_cap_lowest_perf(cap1),
                    amd_cppc_cap_lownonlin_perf(cap1),
                    amd_cppc_cap_nominal_perf(cap1),
                    amd_cppc_cap_highest_perf(cap1),
                ]
                cap_df = pd.concat(
                    [pd.DataFrame([row], columns=cap_df.columns), cap_df],
                    ignore_index=True,
                )

        except FileNotFoundError:
            print_color("Unable to check MSRs: MSR kernel module not loaded", "❌")
            return False
        except PermissionError:
            print_color("MSR checks unavailable", "🚦")
            return

        msr_df = msr_df.sort_values(by="CPU #")
        print_color(
            "CPPC MSRs\n%s"
            % tabulate(msr_df, headers="keys", tablefmt="psql", showindex=False),
            "🔋",
        )

        cap_df = cap_df.sort_values(by="CPU #")
        print_color(
            "MSR_AMD_CPPC_CAP1 (decoded)\n%s"
            % tabulate(cap_df, headers="keys", tablefmt="psql", showindex=False),
            "🔋",
        )

        df = df.sort_values(by="CPU #")
        print_color(
            "MSR_AMD_CPPC_REQ (decoded)\n%s"
            % tabulate(df, headers="keys", tablefmt="psql", showindex=False),
            "🔋",
        )

    def run(self):
        """Run the triage process"""
        self.gather_kernel_info()
        self.gather_amd_pstate_info()
        self.gather_scheduler_info()
        try:
            self.gather_cpu_info()
        except FileNotFoundError:
            print_color("Unable to gather CPU information", "❌")
            return False
        self.gather_msrs()
        return True


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Collect useful information for debugging amd-pstate issues.",
        epilog="Arguments are optional",
    )
    subparsers = parser.add_subparsers(help="Possible commands", dest="command")
    triage_cmd = subparsers.add_parser("triage", help="Run amd-pstate triage")
    triage_cmd.add_argument(
        "--tool-debug",
        action="store_true",
        help="Enable tool debug logging",
    )
    parser.add_argument(
        "--version", action="store_true", help="Show version information"
    )
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    return parser.parse_args()


def main() -> None | int:
    """Main function"""
    args = parse_args()
    ret = False
    if args.version:
        print(version())
        return
    elif args.command == "triage":
        triage = AmdPstateTriage(args.tool_debug)
        ret = triage.run()
    show_log_info()
    if ret is False:
        return 1
    return



#!/usr/bin/python3
# SPDX-License-Identifier: MIT

"""
This module contains unit tests for the battery functions in the amd-debug-tools package.
"""
import unittest
import logging
from unittest.mock import MagicMock, patch


from amd_debug.battery import Batteries


class TestBatteries(unittest.TestCase):
    """Test battery functions"""

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(filename="/dev/null", level=logging.DEBUG)

    @patch("amd_debug.battery.Context")
    def setUp(self, mock_context):
        """Set up a mock context for testing"""
        self.mock_context = mock_context.return_value
        self.batteries = Batteries()

    def test_get_batteries(self):
        """Test getting battery names"""
        mock_device = MagicMock()
        mock_device.device_path = "/devices/LNXSYSTM:00/device:00/PNP0C0A:00"
        mock_device.properties = {"POWER_SUPPLY_NAME": "BAT0"}
        self.mock_context.list_devices.return_value = [mock_device]

        result = self.batteries.get_batteries()
        self.assertEqual(result, ["BAT0"])

    def test_get_energy_unit(self):
        """Test getting energy unit for a battery"""
        mock_device = MagicMock()
        mock_device.device_path = "/devices/LNXSYSTM:00/device:00/PNP0C0A:00"
        mock_device.properties = {
            "POWER_SUPPLY_NAME": "BAT0",
            "POWER_SUPPLY_ENERGY_NOW": "50000",
        }
        self.mock_context.list_devices.return_value = [mock_device]
        result = self.batteries.get_energy_unit("BAT0")
        self.assertEqual(result, "µWh")

    def test_get_energy(self):
        """Test getting current energy for a battery"""
        mock_device = MagicMock()
        mock_device.device_path = "/devices/LNXSYSTM:00/device:00/PNP0C0A:00"
        mock_device.properties = {
            "POWER_SUPPLY_NAME": "BAT0",
            "POWER_SUPPLY_ENERGY_NOW": "50000",
        }
        self.mock_context.list_devices.return_value = [mock_device]
        result = self.batteries.get_energy("BAT0")
        self.assertEqual(result, "50000")

    def test_get_energy_full(self):
        """Test getting full energy for a battery"""
        mock_device = MagicMock()
        mock_device.device_path = "/devices/LNXSYSTM:00/device:00/PNP0C0A:00"
        mock_device.properties = {
            "POWER_SUPPLY_NAME": "BAT0",
            "POWER_SUPPLY_ENERGY_FULL": "60000",
        }
        self.mock_context.list_devices.return_value = [mock_device]

        result = self.batteries.get_energy_full("BAT0")
        self.assertEqual(result, "60000")

    def test_get_description_string(self):
        """Test getting description string for a battery"""
        mock_device = MagicMock()
        mock_device.device_path = "/devices/LNXSYSTM:00/device:00/PNP0C0A:00"
        mock_device.properties = {
            "POWER_SUPPLY_NAME": "BAT0",
            "POWER_SUPPLY_MANUFACTURER": "ACME",
            "POWER_SUPPLY_MODEL_NAME": "SuperBattery",
            "POWER_SUPPLY_ENERGY_FULL": "60000",
            "POWER_SUPPLY_ENERGY_FULL_DESIGN": "80000",
        }
        self.mock_context.list_devices.return_value = [mock_device]

        result = self.batteries.get_description_string("BAT0")
        self.assertEqual(
            result,
            "Battery BAT0 (ACME SuperBattery) is operating at 75.00% of design",
        )



# SPDX-License-Identifier: MIT

"""
This module contains common utility functions and classes for various amd-debug-tools.
"""

import asyncio
import importlib.metadata
import logging
import os
import platform
import time
import struct
import subprocess
import sys
from ast import literal_eval
from datetime import date, timedelta


class Colors:
    """Colors for the terminal"""

    DEBUG = "\033[90m"
    HEADER = "\033[95m"
    OK = "\033[94m"
    WARNING = "\033[32m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    UNDERLINE = "\033[4m"


def read_file(fn) -> str:
    """Read a file and return the contents"""
    with open(fn, "r", encoding="utf-8") as r:
        return r.read().strip()


def compare_file(fn, expect) -> bool:
    """Compare a file to an expected string"""
    return read_file(fn) == expect


def get_group_color(group) -> str:
    """Get the color for a group"""
    if group == "🚦":
        color = Colors.WARNING
    elif group == "🗣️":
        color = Colors.HEADER
    elif any(mk in group for mk in ["💯", "🚫"]):
        color = Colors.UNDERLINE
    elif any(mk in group for mk in ["🦟", "🖴"]):
        color = Colors.DEBUG
    elif any(mk in group for mk in ["❌", "👀"]):
        color = Colors.FAIL
    elif any(mk in group for mk in ["✅", "🔋", "🐧", "💻", "○", "💤", "🥱"]):
        color = Colors.OK
    else:
        color = group
    return color


def print_color(message, group) -> None:
    """Print a message with a color"""
    prefix = f"{group} "
    suffix = Colors.ENDC
    color = get_group_color(group)
    if color == group:
        prefix = ""
    log_txt = f"{prefix}{message}".strip()
    if any(c in color for c in [Colors.OK, Colors.HEADER, Colors.UNDERLINE]):
        logging.info(log_txt)
    elif color == Colors.WARNING:
        logging.warning(log_txt)
    elif color == Colors.FAIL:
        logging.error(log_txt)
    else:
        logging.debug(log_txt)
    if "TERM" in os.environ and os.environ["TERM"] == "dumb":
        suffix = ""
        color = ""
    print(f"{prefix}{color}{message}{suffix}")


def colorize_choices(choices, default) -> str:
    """Output a list of choices with colors, where the default is highlighted"""
    if default not in choices:
        raise ValueError(f"Default choice '{default}' not in choices")
    choices = [c for c in choices if c != default]
    choices = [f"{Colors.OK}{default}{Colors.ENDC}"] + choices
    return ", ".join(choices)


def fatal_error(message):
    """Prints a fatal error message and exits"""
    _configure_log(None)
    print_color(message, "👀")
    sys.exit(1)


def apply_prefix_wrapper(header, message):
    """Apply a prefix to wrap a newline delimitted message"""
    s = f"{header.strip()}\n"
    lines = message.strip().split("\n")
    for i, line in enumerate(lines):
        line = line.strip()
        if not line:
            continue
        if i == len(lines) - 1:
            s += f"└─ {line}\n"
            continue
        s += f"│ {line}\n"
    return s


def show_log_info():
    """Show log information"""
    logger = logging.getLogger()
    for handler in logger.handlers:
        if isinstance(handler, logging.FileHandler):
            filename = handler.baseFilename
            if filename != "/dev/null":
                print(f"Debug logs are saved to: {filename}")


def _configure_log(prefix) -> str:
    """Configure logging for the tool"""
    if len(logging.root.handlers) > 0:
        return

    if prefix:
        user = os.environ.get("SUDO_USER")
        home = os.path.expanduser(f"~{user if user else ''}")
        path = os.environ.get("XDG_DATA_HOME") or os.path.join(
            home, ".local", "share", "amd-debug-tools"
        )
        os.makedirs(path, exist_ok=True)
        log = os.path.join(
            path,
            f"{prefix}-{date.today()}.txt",
        )
        if not os.path.exists(log):
            with open(log, "w", encoding="utf-8") as f:
                f.write("")
            if "SUDO_UID" in os.environ:
                os.chown(path, int(os.environ["SUDO_UID"]), int(os.environ["SUDO_GID"]))
                os.chown(log, int(os.environ["SUDO_UID"]), int(os.environ["SUDO_GID"]))
        level = logging.DEBUG
    else:
        log = "/dev/null"
        level = logging.WARNING
    # for saving a log file for analysis
    logging.basicConfig(
        format="%(asctime)s %(levelname)s:\t%(message)s",
        filename=log,
        level=level,
    )
    return log


def check_lockdown():
    """Check if the system is in lockdown"""
    fn = os.path.join("/", "sys", "kernel", "security", "lockdown")
    if not os.path.exists(fn):
        return False
    lockdown = read_file(fn)
    if lockdown.split()[0] != "[none]":
        return lockdown
    return False


def print_temporary_message(msg) -> int:
    """Print a temporary message to the console"""
    print(msg, end="\r", flush=True)
    return len(msg)


def clear_temporary_message(msg_len) -> None:
    """Clear a temporary message from the console"""
    print(" " * msg_len, end="\r")


def run_countdown(action, t) -> bool:
    """Run a countdown timer"""
    msg = ""
    if t < 0:
        return False
    if t == 0:
        return True
    while t > 0:
        msg = f"{action} in {timedelta(seconds=t)}"
        print_temporary_message(msg)
        time.sleep(1)
        t -= 1
    clear_temporary_message(len(msg))
    return True


def get_distro() -> str:
    """Get the distribution name"""
    distro = "unknown"
    if os.path.exists("/etc/os-release"):
        with open("/etc/os-release", "r", encoding="utf-8") as f:
            for line in f:
                if line.startswith("ID="):
                    return line.split("=")[1].strip().strip('"')
    if os.path.exists("/etc/arch-release"):
        return "arch"
    elif os.path.exists("/etc/fedora-release"):
        return "fedora"
    elif os.path.exists("/etc/debian_version"):
        return "debian"

    return distro


def get_pretty_distro() -> str:
    """Get the pretty distribution name"""
    distro = "Unknown"
    if os.path.exists("/etc/os-release"):
        with open("/etc/os-release", "r", encoding="utf-8") as f:
            for line in f:
                if line.startswith("PRETTY_NAME="):
                    distro = line.split("=")[1].strip().strip('"')
                    break
    return distro


def bytes_to_gb(bytes_value):
    """Convert bytes to GB"""
    return bytes_value * 4096 / (1024 * 1024 * 1024)


def gb_to_pages(gb_value):
    """Convert GB into bytes"""
    return int(gb_value * (1024 * 1024 * 1024) / 4096)


def reboot():
    """Reboot the system"""

    async def reboot_dbus_fast():
        """Reboot using dbus-fast"""
        try:
            from dbus_fast.aio import (  # pylint: disable=import-outside-toplevel
                MessageBus,
            )
            from dbus_fast import BusType  # pylint: disable=import-outside-toplevel

            bus = await MessageBus(bus_type=BusType.SYSTEM).connect()
            introspection = await bus.introspect(
                "org.freedesktop.login1", "/org/freedesktop/login1"
            )
            proxy_obj = bus.get_proxy_object(
                "org.freedesktop.login1", "/org/freedesktop/login1", introspection
            )
            interface = proxy_obj.get_interface("org.freedesktop.login1.Manager")
            await interface.call_reboot(True)

        except ImportError:
            return False
        return True

    def reboot_dbus():
        """Reboot using python-dbus"""
        try:
            import dbus  # pylint: disable=import-outside-toplevel

            bus = dbus.SystemBus()
            obj = bus.get_object("org.freedesktop.login1", "/org/freedesktop/login1")
            intf = dbus.Interface(obj, "org.freedesktop.login1.Manager")
            intf.Reboot(True)
        except ImportError:
            return False
        return True

    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(reboot_dbus_fast())
    if not result:
        return reboot_dbus()

    return True


def get_system_mem():
    """Get the total system memory in GB using /proc/meminfo"""
    with open(os.path.join("/", "proc", "meminfo"), "r", encoding="utf-8") as f:
        for line in f:
            if line.startswith("MemTotal:"):
                # MemTotal line format: "MemTotal:       16384516 kB"
                # Extract the number and convert from kB to GB
                mem_kb = int(line.split()[1])
                return mem_kb / (1024 * 1024)
    raise ValueError("Could not find MemTotal in /proc/meminfo")


def is_root() -> bool:
    """Check if the user is root"""
    return os.geteuid() == 0


def BIT(num):  # pylint: disable=invalid-name
    """Return a bit shifted value"""
    return 1 << num


def get_log_priority(num):
    """Maps an integer debug level to a priority type"""
    if num:
        try:
            num = int(num)
        except ValueError:
            return num
        if num == 7:
            return "🦟"
        elif num == 4:
            return "🚦"
        elif num <= 3:
            return "❌"
    return "○"


def minimum_kernel(major, minor) -> bool:
    """Checks if the kernel version is at least major.minor"""
    ver = platform.uname().release.split(".")
    kmajor = int(ver[0])
    kminor = int(ver[1])
    if kmajor > int(major):
        return True
    if kmajor < int(major):
        return False
    return kminor >= int(minor)


def systemd_in_use() -> bool:
    """Check if systemd is in use"""
    # Check if /proc/1/comm exists and read its contents
    init_daemon = read_file("/proc/1/comm")
    return init_daemon == "systemd"


def get_property_pyudev(properties, key, fallback=""):
    """Get a property from a udev device"""
    try:
        return properties.get(key, fallback)
    except UnicodeDecodeError:
        return ""


def find_ip_version(base_path, kind, hw_ver) -> bool:
    """Determine if an IP version is present on the system"""
    b = os.path.join(base_path, "ip_discovery", "die", "0", kind, "0")
    for key, expected_value in hw_ver.items():
        p = os.path.join(b, key)
        if not os.path.exists(p):
            return False
        v = int(read_file(p))
        if v != expected_value:
            return False
    return True


def read_msr(msr, cpu):
    """Read a Model-Specific Register (MSR) value from the CPU."""
    p = f"/dev/cpu/{cpu}/msr"
    if not os.path.exists(p) and is_root():
        os.system("modprobe msr")
    try:
        f = os.open(p, os.O_RDONLY)
    except OSError as exc:
        raise PermissionError from exc
    try:
        os.lseek(f, msr, os.SEEK_SET)
        val = struct.unpack("Q", os.read(f, 8))[0]
    except OSError as exc:
        raise PermissionError from exc
    finally:
        os.close(f)
    return val


def relaunch_sudo() -> None:
    """Relaunch the script with sudo if not already running as root"""
    if not is_root():
        logging.debug("Relaunching with sudo")
        env_vars = []
        for var in ["DISPLAY", "WAYLAND_DISPLAY", "XAUTHORITY", "XDG_RUNTIME_DIR",
                    "DBUS_SESSION_BUS_ADDRESS", "XDG_SESSION_TYPE"]:
            value = os.environ.get(var)
            if value:
                env_vars.append(f"{var}={value}")

        if env_vars:
            sudo_cmd = ["sudo"] + [f"--preserve-env={var.split('=')[0]}" for var in env_vars] + sys.argv
        else:
            sudo_cmd = ["sudo"] + sys.argv

        os.execvp("sudo", sudo_cmd)


def running_ssh():
    return "SSH_CLIENT" in os.environ or "SSH_TTY" in os.environ


def convert_string_to_bool(str_value) -> bool:
    """convert a string to a boolean value"""
    try:
        value = literal_eval(str_value)
    except (SyntaxError, ValueError):
        value = None
        sys.exit(f"Invalid entry: {str_value}")
    return bool(value)


def _git_describe() -> str:
    """Get the git description of the current commit"""
    try:
        result = subprocess.check_output(
            ["git", "log", "-1", '--format=commit %h ("%s")'],
            cwd=os.path.dirname(__file__),
            text=True,
            stderr=subprocess.DEVNULL,
        )
        return result.strip()
    except subprocess.CalledProcessError:
        return None
    except FileNotFoundError:
        return None


def version() -> str:
    """Get version of the tool"""
    ver = "unknown"
    try:
        ver = importlib.metadata.version("amd-debug-tools")
    except importlib.metadata.PackageNotFoundError:
        pass
    describe = _git_describe()
    if describe:
        ver = f"{ver} [{describe}]"
    return ver


class AmdTool:
    """Base class for AMD tools"""

    def __init__(self, prefix):
        self.log = _configure_log(prefix)
        logging.debug("command: %s (module: %s)", sys.argv, type(self).__name__)
        logging.debug("Version: %s", version())
        if os.uname().sysname != "Linux":
            raise RuntimeError("This tool only runs on Linux")



#!/usr/bin/python3
# SPDX-License-Identifier: MIT

"""
This module contains unit tests for the acpi functions in the amd-debug-tools package.
"""
from unittest.mock import patch, mock_open, call

import logging
import unittest

from amd_debug.acpi import search_acpi_tables, AcpicaTracer, ACPI_METHOD


class TestAcpi(unittest.TestCase):
    """Test acpi functions"""

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(filename="/dev/null", level=logging.DEBUG)

    def test_search_acpi_tables(self):
        """Test search_acpi_tables function"""
        pattern = "test_pattern"
        bad_pattern = "bad_pattern"
        mock_listdir = ["ABA", "SSDT1", "DSDT2", "SSDT3"]
        mock_file_content = b"test_pattern"

        with patch("os.listdir", return_value=mock_listdir), patch(
            "builtins.open", mock_open(read_data=mock_file_content)
        ):
            result = search_acpi_tables(pattern)
            self.assertTrue(result)

        with patch("os.listdir", return_value=mock_listdir), patch(
            "builtins.open", mock_open(read_data=mock_file_content)
        ):
            result = search_acpi_tables(bad_pattern)
            self.assertFalse(result)

        with patch("os.listdir", return_value=["OTHER1", "OTHER2"]), patch(
            "builtins.open", mock_open(read_data=b"no_match")
        ):
            result = search_acpi_tables(pattern)
            self.assertFalse(result)

    def test_acpica_tracer_missing_bios(self):
        """Test AcpicaTracer class when ACPI tracing is not supported"""

        mock_listdir = ["SSDT1", "DSDT2", "SSDT3"]

        with patch("os.listdir", return_value=mock_listdir), patch(
            "builtins.open", mock_open(read_data=b"foo")
        ), patch("os.path.exists", return_value=True):

            tracer = AcpicaTracer()
            self.assertTrue(tracer.supported)

            self.assertFalse(tracer.trace_bios())

    def test_acpica_tracer(self):
        """Test AcpicaTracer class"""

        mock_listdir = ["SSDT1", "DSDT2", "SSDT3"]
        mock_file_content = bytes(ACPI_METHOD, "utf-8")

        with patch("os.listdir", return_value=mock_listdir), patch(
            "builtins.open", mock_open(read_data=mock_file_content)
        ), patch("os.path.exists", return_value=True):

            tracer = AcpicaTracer()
            self.assertTrue(tracer.supported)

            self.assertTrue(tracer.trace_notify())
            self.assertTrue(tracer.trace_bios())
            self.assertTrue(tracer.disable())
            self.assertTrue(tracer.restore())

    def test_acpica_trace_no_acpi_debug(self):
        """Test AcpicaTracer class when ACPI tracing is not supported"""
        with patch("os.path.exists", return_value=False), patch(
            "amd_debug.common.open", mock_open(read_data="foo")
        ):
            tracer = AcpicaTracer()
            self.assertFalse(tracer.supported)

            self.assertFalse(tracer.trace_notify())
            self.assertFalse(tracer.trace_bios())
            self.assertFalse(tracer.disable())
            self.assertFalse(tracer.restore())



# SPDX-License-Identifier: MIT
"""s2idle analysis tool"""
import argparse
import sys
import os
import subprocess
import sqlite3

from datetime import date, timedelta, datetime
from amd_debug.common import (
    convert_string_to_bool,
    colorize_choices,
    is_root,
    relaunch_sudo,
    show_log_info,
    version,
    running_ssh,
)

from amd_debug.validator import SleepValidator
from amd_debug.installer import Installer
from amd_debug.prerequisites import PrerequisiteValidator
from amd_debug.sleep_report import SleepReport


class Defaults:
    """Default values for the script"""

    duration = 10
    wait = 4
    count = 1
    since = date.today() - timedelta(days=60)
    until = date.today() + timedelta(days=1)
    format_choices = ["txt", "md", "html", "stdout"]
    boolean_choices = ["true", "false"]


class Headers:
    """Headers for the script"""

    DurationDescription = "How long should suspend cycles last (seconds)"
    WaitDescription = "How long to wait in between suspend cycles (seconds)"
    CountDescription = "How many suspend cycles to run"
    SinceDescription = "What date to start report data"
    UntilDescription = "What date to end report data"
    LogDescription = "Location of log file"
    ReportFileDescription = "Location of report file"
    FormatDescription = "What format to output the report in"
    MaxDurationDescription = "What is the maximum suspend cycle length (seconds)"
    MaxWaitDescription = "What is the maximum time between suspend cycles (seconds)"
    ReportDebugDescription = "Enable debug output in report (increased size)"


def display_report_file(fname, fmt) -> None:
    """Display report file"""
    if fmt != "html":
        return
    if not is_root():
        subprocess.call(["xdg-open", fname])
        return
    user = os.environ.get("SUDO_USER")
    if user:
        env_vars = []
        for var in ["DISPLAY", "WAYLAND_DISPLAY", "XAUTHORITY", "XDG_RUNTIME_DIR",
                    "DBUS_SESSION_BUS_ADDRESS", "XDG_SESSION_TYPE"]:
            value = os.environ.get(var)
            if value:
                env_vars.append(f"{var}={value}")

        if env_vars:
            cmd = ["sudo", "-u", user] + [f"env"] + env_vars + ["xdg-open", fname]
            subprocess.call(cmd)
        else:
            print(
                "Unable to detect graphical session environment. "
                "Report saved to: " + fname
            )


def get_report_file(report_file, extension) -> str:
    """Prompt user for report file"""
    if extension == "stdout":
        return ""
    if not report_file:
        return f"amd-s2idle-report-{date.today()}.{extension}"
    return report_file


def get_report_format() -> str:
    """Get report format"""
    if running_ssh():
        return "txt"
    return "html"


def prompt_report_arguments(since, until, fname, fmt, report_debug) -> list:
    """Prompt user for report configuration"""
    if not since:
        default = Defaults.since
        since = input(f"{Headers.SinceDescription} (default {default})? ")
        if not since:
            since = default.isoformat()
    try:
        since = datetime.fromisoformat(since)
    except ValueError as e:
        sys.exit(f"Invalid date, use YYYY-MM-DD: {e}")
    if not until:
        default = Defaults.until
        until = input(f"{Headers.SinceDescription} (default {default})? ")
        if not until:
            until = default.isoformat()
    try:
        until = datetime.fromisoformat(until)
    except ValueError as e:
        sys.exit(f"Invalid date, use YYYY-MM-DD: {e}")

    if not fmt:
        fmt = input(
            f"{Headers.FormatDescription} ({colorize_choices(Defaults.format_choices, get_report_format())})? "
        )
        if not fmt:
            fmt = get_report_format()
        if fmt not in Defaults.format_choices:
            sys.exit(f"Invalid format: {fmt}")
    if report_debug is None:
        inp = (
            input(
                f"{Headers.ReportDebugDescription} ({colorize_choices(Defaults.boolean_choices, 'true')})? "
            )
            .lower()
            .capitalize()
        )
        report_debug = True if not inp else convert_string_to_bool(inp)
    return [since, until, get_report_file(fname, fmt), fmt, report_debug]


def prompt_test_arguments(duration, wait, count, rand) -> list:
    """Prompt user for test configuration"""
    if not duration:
        if rand:
            question = Headers.MaxDurationDescription
        else:
            question = Headers.DurationDescription
        duration = input(f"{question} (default {Defaults.duration})? ")
        if not duration:
            duration = Defaults.duration
    try:
        duration = int(duration)
    except ValueError as e:
        sys.exit(f"Invalid duration: {e}")
    if not wait:
        if rand:
            question = Headers.MaxWaitDescription
        else:
            question = Headers.WaitDescription
        wait = input(f"{question} (default {Defaults.wait})? ")
        if not wait:
            wait = Defaults.wait
    try:
        wait = int(wait)
    except ValueError as e:
        sys.exit(f"Invalid wait: {e}")
    if not count:
        count = input(f"{Headers.CountDescription} (default {Defaults.count})? ")
        if not count:
            count = Defaults.count
    try:
        count = int(count)
    except ValueError as e:
        sys.exit(f"Invalid count: {e}")
    return [duration, wait, count]


def report(since, until, fname, fmt, tool_debug, report_debug) -> bool:
    """Generate a report from previous sleep cycles"""
    try:
        since, until, fname, fmt, report_debug = prompt_report_arguments(
            since, until, fname, fmt, report_debug
        )
    except KeyboardInterrupt:
        sys.exit("\nReport generation cancelled")
    try:
        app = SleepReport(
            since=since,
            until=until,
            fname=fname,
            fmt=fmt,
            tool_debug=tool_debug,
            report_debug=report_debug,
        )
    except sqlite3.OperationalError as e:
        print(f"Failed to generate report: {e}")
        return False
    except PermissionError as e:
        print(f"Failed to generate report: {e}")
        return False
    try:
        app.run()
    except PermissionError as e:
        print(f"Failed to generate report: {e}")
        return False
    except ValueError as e:
        print(f"Failed to generate report: {e}")
        return False
    display_report_file(fname, fmt)
    return True


def run_test_cycle(
    duration, wait, count, fmt, fname, force, debug, rand, logind, bios_debug
) -> bool:
    """Run a test"""
    app = Installer(tool_debug=debug)
    app.set_requirements("iasl", "ethtool", "edid-decode")
    if not app.install_dependencies():
        print("Failed to install dependencies")
        return False

    try:
        duration, wait, count = prompt_test_arguments(duration, wait, count, rand)
        total_seconds = (duration + wait) * count
        until_time = datetime.now() + timedelta(seconds=total_seconds)
        since, until, fname, fmt, report_debug = prompt_report_arguments(
            datetime.now().isoformat(), until_time.isoformat(), fname, fmt, True
        )
    except KeyboardInterrupt:
        sys.exit("\nTest cancelled")

    try:
        app = PrerequisiteValidator(debug)
        run = app.run()
    except PermissionError as e:
        print(f"Failed to run prerequisite check: {e}")
        return False
    app.report()

    if run or force:
        app = SleepValidator(tool_debug=debug, bios_debug=bios_debug)

        run = app.run(
            duration=duration,
            wait=wait,
            count=count,
            rand=rand,
            logind=logind,
        )
        until = datetime.now()
    else:
        since = None
        until = None

    app = SleepReport(
        since=since,
        until=until,
        fname=fname,
        fmt=fmt,
        tool_debug=debug,
        report_debug=report_debug,
    )
    app.run()

    # open report in browser if it's html
    display_report_file(fname, fmt)

    return True


def install(debug) -> None:
    """Install the tool"""
    installer = Installer(tool_debug=debug)
    installer.set_requirements("iasl", "ethtool", "edid-decode")
    if not installer.install_dependencies():
        sys.exit("Failed to install dependencies")
    try:
        app = PrerequisiteValidator(debug)
        run = app.run()
    except PermissionError as e:
        sys.exit(f"Failed to run prerequisite check: {e}")
    if not run:
        app.report()
        sys.exit("Failed to meet prerequisites")

    if not installer.install():
        sys.exit("Failed to install")


def uninstall(debug) -> None:
    """Uninstall the tool"""
    app = Installer(tool_debug=debug)
    if not app.remove():
        sys.exit("Failed to remove")


def parse_args():
    """Parse command line arguments"""
    parser = argparse.ArgumentParser(
        description="Swiss army knife for analyzing Linux s2idle problems",
        epilog="The tool can run an immediate test with the 'test' command "
        "or can be used to hook into systemd for building reports later.\n"
        "All optional arguments will be prompted if needed.\n"
        "To use non-interactively, please populate all optional arguments.",
    )
    subparsers = parser.add_subparsers(help="Possible commands", dest="action")

    # 'test' command
    test_cmd = subparsers.add_parser("test", help="Run amd-s2idle test and report")
    test_cmd.add_argument("--count", help=Headers.CountDescription)
    test_cmd.add_argument(
        "--duration",
        help=Headers.DurationDescription,
    )
    test_cmd.add_argument(
        "--wait",
        help=Headers.WaitDescription,
    )
    test_cmd.add_argument(
        "--logind", action="store_true", help="Use logind to suspend system"
    )
    test_cmd.add_argument(
        "--random",
        action="store_true",
        help="Run sleep cycles for random durations and wait, using the "
        "--duration and --wait arguments as an upper bound",
    )
    test_cmd.add_argument(
        "--force",
        action="store_true",
        help="Run suspend test even if prerequisites failed",
    )
    test_cmd.add_argument(
        "--format",
        choices=Defaults.format_choices,
        help="Report format",
    )
    test_cmd.add_argument(
        "--tool-debug",
        action="store_true",
        help="Enable tool debug logging",
    )
    test_cmd.add_argument(
        "--bios-debug",
        action="store_true",
        help="Enable BIOS debug logging instead of notify logging",
    )
    test_cmd.add_argument("--report-file", help=Headers.ReportFileDescription)

    # 'report' command
    report_cmd = subparsers.add_parser(
        "report", help="Generate amd-s2idle report from previous runs"
    )
    report_cmd.add_argument(
        "--since",
        help=Headers.SinceDescription,
    )
    report_cmd.add_argument(
        "--until",
        default=Defaults.until.isoformat(),
        help=Headers.UntilDescription,
    )
    report_cmd.add_argument("--report-file", help=Headers.ReportFileDescription)
    report_cmd.add_argument(
        "--format",
        choices=Defaults.format_choices,
        help="Report format",
    )
    report_cmd.add_argument(
        "--tool-debug",
        action="store_true",
        help="Enable tool debug logging",
    )
    report_cmd.add_argument(
        "--report-debug",
        action=argparse.BooleanOptionalAction,
        help="Include debug messages in report (WARNING: can significantly increase report size)",
    )

    # if running in a venv, install/uninstall hook options
    if sys.prefix != sys.base_prefix:
        install_cmd = subparsers.add_parser(
            "install", help="Install systemd s2idle hook"
        )
        uninstall_cmd = subparsers.add_parser(
            "uninstall", help="Uninstall systemd s2idle hook"
        )
        install_cmd.add_argument(
            "--tool-debug",
            action="store_true",
            help="Enable tool debug logging",
        )
        uninstall_cmd.add_argument(
            "--tool-debug",
            action="store_true",
            help="Enable tool debug logging",
        )

    parser.add_argument(
        "--version", action="store_true", help="Show version information"
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return parser.parse_args()


def main() -> None | int:
    """Main function"""
    args = parse_args()
    ret = False
    if args.action == "install":
        relaunch_sudo()
        install(args.tool_debug)
    elif args.action == "uninstall":
        relaunch_sudo()
        uninstall(args.tool_debug)
    elif args.action == "report":
        ret = report(
            args.since,
            args.until,
            args.report_file,
            args.format,
            args.tool_debug,
            args.report_debug,
        )
    elif args.action == "test":
        relaunch_sudo()
        ret = run_test_cycle(
            args.duration,
            args.wait,
            args.count,
            args.format,
            args.report_file,
            args.force,
            args.tool_debug,
            args.random,
            args.logind,
            args.bios_debug,
        )
    elif args.version:
        print(version())
        return
    else:
        sys.exit("no action specified")
    show_log_info()
    if ret is False:
        return 1
    return



#!/usr/bin/python3
# SPDX-License-Identifier: MIT

"""
This module contains unit tests for the installer functions in the amd-debug-tools package.
"""
from unittest.mock import patch, mock_open

import logging
import unittest

from amd_debug.installer import Installer


class TestInstaller(unittest.TestCase):
    """Test installer functions"""

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(filename="/dev/null", level=logging.DEBUG)

    def setUp(self):
        self.installer = Installer(tool_debug=False)

    @patch("builtins.print")
    @patch("shutil.copy", return_value=None)
    @patch("os.chmod", return_value=None)
    @patch("builtins.open")
    @patch("subprocess.call", return_value=0)
    @patch("os.makedirs", return_value=None)
    def test_install_hook(
        self,
        _mock_mkdir,
        _mock_call,
        _mock_open,
        _mock_chmod,
        _mock_shutil,
        _mock_print,
    ):
        """Test install hook function"""
        self.installer.install()

    @patch("builtins.print")
    @patch("shutil.copy", return_value=None)
    @patch("os.chmod", return_value=None)
    @patch("builtins.open")
    @patch("subprocess.call", return_value=0)
    @patch("os.makedirs", return_value=None)
    def test_install_hook_missing_path(
        self,
        _mock_mkdir,
        _mock_call,
        _mock_open,
        _mock_chmod,
        _mock_shutil,
        _mock_print,
    ):
        """Test install hook function, but some paths are missing"""
        with patch("os.path.exists", return_value=False):
            self.installer.install()

    @patch("builtins.print")
    @patch("os.remove", return_value=None)
    @patch("builtins.open")
    @patch("amd_debug.installer.get_distro", return_value="ubuntu")
    @patch("subprocess.call", return_value=0)
    def test_remove_hook(
        self, _mock_call, _mock_get_distro, _mock_open, _mock_remove, _mock_print
    ):
        """Test remove hook function"""
        with patch("os.path.exists", return_value=True):
            self.installer.remove()

    @patch("builtins.print")
    @patch("os.path.exists", return_value=False)
    @patch("subprocess.call", return_value=0)
    def test_remove_hook_missing_path(self, _mock_call, _mock_exists, _mock_print):
        """Test remove hook function when the file is missing"""
        self.installer.remove()

    @patch("builtins.print")
    @patch("subprocess.call", return_value=0)
    def test_already_installed_iasl(self, _mock_call, _mock_print):
        """Test that an already installed iasl is found"""
        self.installer.set_requirements("iasl")
        ret = self.installer.install_dependencies()
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="ubuntu")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_iasl_ubuntu(
        self, _mock_call, _mock_check_call, _mock_distro, _fake_sudo, _mock_print
    ):
        """Test install requirements function"""
        self.installer.set_requirements("iasl")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(["apt", "install", "acpica-tools"])
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="VARIANT_ID=workstation\n"
    )
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_iasl_fedora(
        self,
        _mock_call,
        _mock_check_call,
        _mock_variant,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function"""
        self.installer.set_requirements("iasl")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(
            ["dnf", "install", "-y", "acpica-tools"]
        )
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch("builtins.open", new_callable=mock_open, read_data="VARIANT_ID=kde\n")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_iasl_fedora_kde(
        self,
        _mock_call,
        _mock_check_call,
        _mock_variant,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function on Fedora KDE"""
        self.installer.set_requirements("iasl")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(
            ["dnf", "install", "-y", "acpica-tools"]
        )
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="ubuntu")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_ethtool_ubuntu(
        self, _mock_call, _mock_check_call, _mock_distro, _fake_sudo, _mock_print
    ):
        """Test install requirements function"""
        self.installer.set_requirements("ethtool")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(["apt", "install", "ethtool"])
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="VARIANT_ID=workstation\n"
    )
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_ethtool_fedora(
        self,
        _mock_call,
        _mock_check_call,
        _mock_variant,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function"""
        self.installer.set_requirements("ethtool")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(["dnf", "install", "-y", "ethtool"])
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch("builtins.open", new_callable=mock_open, read_data="VARIANT_ID=kde\n")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_ethtool_fedora_kde(
        self,
        _mock_call,
        _mock_check_call,
        _mock_variant,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function on Fedora KDE"""
        self.installer.set_requirements("ethtool")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(["dnf", "install", "-y", "ethtool"])
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="arch")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_ethtool_arch(
        self,
        _mock_call,
        _mock_check_call,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function"""
        self.installer.set_requirements("ethtool")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(["pacman", "-Sy", "ethtool"])
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("os.path.exists", return_value=False)
    @patch("os.execvp", return_value=None)
    @patch("amd_debug.installer.get_distro", return_value="gentoo")
    @patch("subprocess.call", return_value=1)
    def test_install_iasl_gentoo(
        self, _mock_call, _mock_distro, _fake_sudo, _mock_exists, _mock_print
    ):
        """Test install requirements function"""
        self.installer.set_requirements("iasl", "ethtool")
        ret = self.installer.install_dependencies()
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="ubuntu")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_edid_decode_ubuntu(
        self, _mock_call, _mock_check_call, _mock_distro, _fake_sudo, _mock_print
    ):
        """Test install requirements function for edid-decode on Ubuntu"""
        self.installer.set_requirements("edid-decode")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(
            ["apt", "install", "libdisplay-info-bin"]
        )
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="VARIANT_ID=workstation\n"
    )
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_edid_decode_fedora(
        self,
        _mock_call,
        _mock_check_call,
        _mock_variant,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function for edid-decode on Fedora"""
        self.installer.set_requirements("edid-decode")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(
            ["dnf", "install", "-y", "libdisplay-info-tools"]
        )
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch("builtins.open", new_callable=mock_open, read_data="VARIANT_ID=kde\n")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_edid_decode_fedora_kde(
        self,
        _mock_call,
        _mock_check_call,
        _mock_variant,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function for edid-decode on Fedora KDE"""
        self.installer.set_requirements("edid-decode")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(
            ["dnf", "install", "-y", "libdisplay-info-tools"]
        )
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="arch")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_edid_decode_arch(
        self,
        _mock_call,
        _mock_check_call,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function for edid-decode on Arch"""
        self.installer.set_requirements("edid-decode")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(["pacman", "-Sy", "libdisplay-info"])
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("os.path.exists", return_value=False)
    @patch("os.execvp", return_value=None)
    @patch("amd_debug.installer.get_distro", return_value="gentoo")
    @patch("subprocess.call", return_value=1)
    def test_install_edid_decode_gentoo(
        self, _mock_call, _mock_distro, _fake_sudo, _mock_exists, _mock_print
    ):
        """Test install requirements function for edid-decode on unsupported distro"""
        self.installer.set_requirements("edid-decode")
        ret = self.installer.install_dependencies()
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("os.path.exists", return_value=False)
    @patch("os.execvp", return_value=None)
    @patch("amd_debug.installer.get_distro", return_value="gentoo")
    @patch("subprocess.call", return_value=255)
    def test_install_edid_decode_present(
        self, _mock_call, _mock_distro, _fake_sudo, _mock_exists, _mock_print
    ):
        """Test install requirements function for edid-decode on unsupported distro"""
        self.installer.set_requirements("edid-decode")
        ret = self.installer.install_dependencies()
        self.assertTrue(ret)



# SPDX-License-Identifier: MIT
"""Kernel log analysis"""

import logging
import re
import os
import subprocess
from datetime import timedelta

from amd_debug.common import systemd_in_use, read_file, fatal_error


def get_kernel_command_line() -> str:
    """Get the kernel command line"""
    cmdline = read_file(os.path.join("/proc", "cmdline"))
    # borrowed from https://github.com/fwupd/fwupd/blob/1.9.5/libfwupdplugin/fu-common-linux.c#L95
    filtered = [
        "apparmor",
        "audit",
        "auto",
        "bluetooth.disable_ertm",
        "boot",
        "BOOT_IMAGE",
        "console",
        "crashkernel",
        "cryptdevice",
        "cryptkey",
        "dm",
        "earlycon",
        "earlyprintk",
        "ether",
        "init",
        "initrd",
        "ip",
        "LANG",
        "loglevel",
        "luks.key",
        "luks.name",
        "luks.options",
        "luks.uuid",
        "mitigations",
        "mount.usr",
        "mount.usrflags",
        "mount.usrfstype",
        "netdev",
        "netroot",
        "nfsaddrs",
        "nfs.nfs4_unique_id",
        "nfsroot",
        "noplymouth",
        "nowatchdog",
        "ostree",
        "preempt",
        "quiet",
        "rd.dm.uuid",
        "rd.luks.allow-discards",
        "rd.luks.key",
        "rd.luks.name",
        "rd.luks.options",
        "rd.luks.uuid",
        "rd.lvm.lv",
        "rd.lvm.vg",
        "rd.md.uuid",
        "rd.systemd.mask",
        "rd.systemd.wants",
        "resume",
        "resumeflags",
        "rhgb",
        "ro",
        "root",
        "rootflags",
        "rootfstype",
        "roothash",
        "rw",
        "security",
        "selinux",
        "showopts",
        "splash",
        "swap",
        "systemd.machine_id",
        "systemd.mask",
        "systemd.show_status",
        "systemd.unit",
        "systemd.verity_root_data",
        "systemd.verity_root_hash",
        "systemd.wants",
        "udev.log_priority",
        "verbose",
        "vt.handoff",
        "zfs",
        "zswap.enabled",
    ]
    # remove anything that starts with something in filtered from cmdline
    return " ".join([x for x in cmdline.split() if not x.startswith(tuple(filtered))])


def sscanf_bios_args(line):
    """Extracts the format string and arguments from a BIOS trace line"""
    if re.search(r"ex_trace_point", line):
        return True
    elif re.search(r"ex_trace_args", line):
        parts = line.split(": ", 1)
        if len(parts) < 2:
            return None

        t = parts[1].strip()
        match = re.match(r'"(.*?)"(,.*)', t)
        if match:
            format_string = match.group(1).strip().replace("\\n", "")
            args_part = match.group(2).strip(", ")
            arguments = [arg.strip() for arg in args_part.split(",")]

            format_specifiers = re.findall(r"%([xXdD])", format_string)

            converted_args = []
            arg_index = 0
            for _specifier in format_specifiers:
                if arg_index < len(arguments):
                    value = arguments[arg_index]
                    if value == "Unknown":
                        converted_args.append(-1)
                    else:
                        try:
                            converted_args.append(int(value, 16))
                        except ValueError:
                            return None
                    arg_index += 1
                else:
                    break

            try:
                return format_string % tuple(converted_args)
            except TypeError:
                return None
        else:
            # If no format string is found, assume no format modifiers and return True
            return True
    # evmisc-0132 ev_queue_notify_reques: Dispatching Notify on [UBTC] (Device) Value 0x80 (Status Change) Node 00000000851b15c1
    elif re.search(r"ev_queue_notify_reques", line):
        parts = line.split(": ", 1)
        if len(parts) < 2:
            return None
        return parts[1].split("Node")[0].strip()

    return None


class KernelLogger:
    """Base class for kernel loggers"""

    def __init__(self):
        pass

    def seek(self):
        """Seek to the beginning of the log"""

    def seek_tail(self, tim=None):
        """Seek to the end of the log"""

    def process_callback(self, callback, priority=None):
        """Process the log"""

    def match_line(self, _matches) -> str:
        """Find lines that match all matches"""
        return ""

    def match_pattern(self, _pattern) -> str:
        """Find lines that match a pattern"""
        return ""

    def get_full_log(self) -> str:
        """Get the full log as a string"""
        return ""


class InputFile(KernelLogger):
    """Class for input file parsing"""

    def __init__(self, fname):
        self.since_support = False
        self.buffer = ""
        self.seeked = False
        self.buffer = read_file(fname)

    def process_callback(self, callback, priority=None):
        """Process the log"""
        for entry in self.buffer.split("\n"):
            callback(entry, priority)

    def get_full_log(self):
        """Get the full log as a string"""
        return self.buffer


class DmesgLogger(KernelLogger):
    """Class for dmesg logging"""

    def __init__(self):
        self.since_support = False
        self.buffer = ""
        self.seeked = False

        cmd = ["dmesg", "-h"]
        result = subprocess.run(cmd, check=True, capture_output=True)
        for line in result.stdout.decode("utf-8").split("\n"):
            if "--since" in line:
                self.since_support = True
        logging.debug("dmesg since support: %d", self.since_support)

        self.command = ["dmesg", "-t", "-k"]
        self._refresh_head()

    def _refresh_head(self):
        self.buffer = ""
        self.seeked = False
        result = subprocess.run(self.command, check=True, capture_output=True)
        if result.returncode == 0:
            self.buffer = result.stdout.decode("utf-8")

    def seek(self):
        """Seek to the beginning of the log"""
        if self.seeked:
            self._refresh_head()

    def seek_tail(self, tim=None):
        """Seek to the end of the log"""
        if tim:
            if self.since_support:
                # look 10 seconds back because dmesg time isn't always accurate
                fuzz = tim - timedelta(seconds=10)
                cmd = self.command + [
                    "--time-format=iso",
                    f"--since={fuzz.strftime('%Y-%m-%dT%H:%M:%S')}",
                ]
            else:
                cmd = self.command
            result = subprocess.run(cmd, check=True, capture_output=True)
            if result.returncode == 0:
                self.buffer = result.stdout.decode("utf-8")
                if self.since_support:
                    self.seeked = True

    def process_callback(self, callback, _priority=None):
        """Process the log"""
        for entry in self.buffer.split("\n"):
            callback(entry, _priority)

    def match_line(self, matches):
        """Find lines that match all matches"""
        for entry in self.buffer.split("\n"):
            for match in matches:
                if match not in entry:
                    break
                return entry
        return ""

    def match_pattern(self, pattern) -> str:
        for entry in self.buffer.split("\n"):
            if re.search(pattern, entry):
                return entry
        return ""

    def capture_header(self):
        """Capture the header of the log"""
        return self.buffer.split("\n")[0]

    def get_full_log(self):
        """Get the full log as a string"""
        return self.buffer


class CySystemdLogger(KernelLogger):
    """Class for logging using systemd journal using cython"""

    def __init__(self):
        from cysystemd.reader import (
            JournalReader,
            JournalOpenMode,
            Rule,
        )  # pylint: disable=import-outside-toplevel

        boot_reader = JournalReader()
        boot_reader.open(JournalOpenMode.SYSTEM)
        boot_reader.seek_tail()
        boot_reader.skip_previous(1)

        current_boot_id = None
        for entry in boot_reader:
            if hasattr(entry, "data") and "_BOOT_ID" in entry.data:
                current_boot_id = entry.data["_BOOT_ID"]
                break
        if not current_boot_id:
            raise RuntimeError("Unable to find current boot ID")

        rules = Rule("_BOOT_ID", current_boot_id) & Rule("_TRANSPORT", "kernel")

        self.journal = JournalReader()
        self.journal.open(JournalOpenMode.SYSTEM)
        self.journal.add_filter(rules)

    def seek(self):
        """Seek to the beginning of the log"""
        self.journal.seek_head()

    def seek_tail(self, tim=None):
        """Seek to the end of the log"""
        if tim:
            timestamp_usec = int(tim.timestamp() * 1_000_000)
            self.journal.seek_realtime_usec(timestamp_usec)
        else:
            self.journal.seek_tail()

    def process_callback(self, callback, _priority=None):
        """Process the log"""
        for entry in self.journal:
            callback(entry["MESSAGE"], entry["PRIORITY"])

    def match_line(self, matches):
        """Find lines that match all matches"""
        for entry in self.journal:
            for match in matches:
                if match not in entry["MESSAGE"]:
                    break
                return entry["MESSAGE"]
        return None

    def match_pattern(self, pattern):
        """Find lines that match a pattern"""
        for entry in self.journal:
            if re.search(pattern, entry["MESSAGE"]):
                return entry["MESSAGE"]
        return None

    def get_full_log(self):
        """Get the full log as a string"""
        self.seek()
        return "\n".join([entry["MESSAGE"] for entry in self.journal])


class SystemdLogger(KernelLogger):
    """Class for logging using systemd journal"""

    def __init__(self):
        from systemd import journal  # pylint: disable=import-outside-toplevel

        self.journal = journal.Reader()
        self.journal.this_boot()
        self.journal.log_level(journal.LOG_INFO)
        self.journal.add_match(_TRANSPORT="kernel")
        self.journal.add_match(PRIORITY=journal.LOG_DEBUG)

    def seek(self):
        """Seek to the beginning of the log"""
        self.journal.seek_head()

    def seek_tail(self, tim=None):
        if tim:
            self.journal.seek_realtime(tim)
        else:
            self.journal.seek_tail()

    def process_callback(self, callback, _priority=None):
        """Process the log"""
        for entry in self.journal:
            callback(entry["MESSAGE"], entry["PRIORITY"])

    def match_line(self, matches):
        """Find lines that match all matches"""
        for entry in self.journal:
            for match in matches:
                if match not in entry["MESSAGE"]:
                    break
                return entry["MESSAGE"]
        return ""

    def match_pattern(self, pattern):
        """Find lines that match a pattern"""
        for entry in self.journal:
            if re.search(pattern, entry["MESSAGE"]):
                return entry["MESSAGE"]
        return ""

    def get_full_log(self):
        """Get the full log as a string"""
        self.seek()
        return "\n".join([entry["MESSAGE"] for entry in self.journal])


def get_kernel_log(input_file=None) -> KernelLogger:
    """Get the kernel log provider"""
    kernel_log = None
    if input_file:
        kernel_log = InputFile(input_file)
    elif systemd_in_use():
        try:
            kernel_log = CySystemdLogger()
        except ImportError:
            kernel_log = None
        except RuntimeError as e:
            logging.debug(e)
            kernel_log = None
        if not kernel_log:
            try:
                kernel_log = SystemdLogger()
            except ModuleNotFoundError:
                pass
    if not kernel_log:
        try:
            kernel_log = DmesgLogger()
        except subprocess.CalledProcessError as e:
            fatal_error(f"{e}")
            kernel_log = KernelLogger()
    logging.debug("Kernel log provider: %s", kernel_log.__class__.__name__)
    return kernel_log
